from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import re
from itertools import chain
from xml.etree import ElementTree as ET

from .read_label import read_label

from ..utils.constants import PDS4_NAMESPACES
from ..utils.helpers import xml_to_dict, find_xml_element

from ..extern import six


class Label(object):
    """ Stores a PDS4 label or a portion of a PDS4 label.

        This class is similar to an ``ElementTree`` Element; most of the basic attributes and methods
        for Element will also work here. Some additional convenience methods are provided to make it
        easier to work with PDS4 labels.

        Parameters
        ----------
        (in general `load` should be used instead of using __init__)

        convenient_root : ``ElementTree`` Element
            Root element of the PDS4 label or label portion, modified to strip extra whitespaces
            in all element and attribute values and to use the default PDS4 prefixes.
        unmodified_root : ``ElementTree`` Element
            Root element of the PDS4 label or label portion.
        namespace_map: dict
            A `dict` with keys being the namespace URIs and values being the namespace prefixes for
            this label.

        Examples
        --------

        To load a label from file,

        >>> lbl = Label()
        >>> lbl.load('/path/to/label.xml')

        To search the entire label for logical_identifier (of which only one is allowed),

        >>> lbl.find('.//logical_identifier)

        To search the entire label for Display Settings in the 'disp' namespace
        (of which there could be multiple),

        >>> display_settings = lbl.findall('.//disp:Display_Settings')

        To search the Display Settings for the top-level element, Display_Direction,

        >>> display_settings.find('disp:Display_Direction')

        See method descriptions for more info.
    """

    def __init__(self, convenient_root=None, unmodified_root=None, namespace_map=None):

        # ElementTree root containing read-in XML which was modified before storage
        # (see options set in load() below)
        self._convenient_root = convenient_root

        # ElementTree root containing the unmodified XML
        self._unmodified_root = unmodified_root

        # A dictionary with keys being the namespace and values being the namespace prefix for this label
        self._namespace_map = namespace_map

    def __getitem__(self, key):
        """ Obtain subelement.

        Parameters
        ----------
        key : int, or slice
            Specifies which subelement to select

        Returns
        -------
        Label
            Subelement of label

        """
        return Label(self._convenient_root[key], self._unmodified_root[key], namespace_map=self._namespace_map)

    def __repr__(self):
        """
        Returns
        -------
        str
            A repr string identifying the tag element in a similar way to ``ElementTree`` Element.
        """
        if self._convenient_root is None:
            return super(Label, self).__repr__()
        else:
            return '<Element {0} at {1}>'.format(self._convenient_root.tag, hex(id(self)))

    def load(self, filename):
        """ Load Label from PDS4 XML label.

        Parameters
        ----------
        filename : str
            Filename, including path, to XML label.

        Returns
        -------
        None

        """

        self._convenient_root, self._namespace_map = read_label(filename, strip_extra_whitespace=True,
                                           enforce_default_prefixes=True, include_namespace_map=True)

        self._unmodified_root = read_label(filename, strip_extra_whitespace=False,
                                           enforce_default_prefixes=False, include_namespace_map=False)

    @property
    def text(self):
        """
        Returns
        -------
            Text of element.

        """
        return self.getroot().text

    @property
    def tail(self):
        """
        Returns
        -------
            Tail of element.
        """
        return self.getroot().tail

    @property
    def tag(self):
        """
        Returns
        -------
            Tag of element.
        """
        return self.getroot().tag

    @property
    def attrib(self):
        """
        Returns
        -------
            Attributes of element.
        """
        return self.getroot().attrib

    def getroot(self, unmodified=False):
        """
        Parameters
        ----------
        unmodified : bool, optional
            If True, returns the unmodified ``ElementTree`` Element. Defaults to False.

        Returns
        -------
        Element instance
            The ``ElementTree`` root element backing this Label.
        """

        root = self._convenient_root
        if unmodified:
            root = self._unmodified_root

        return root

    def find(self, match, namespaces=None, unmodified=False, return_ET=False):
        """ Search for the first matching subelement.

        Uses the same format as ``ElementTree.find``. See `Label` docstring or ``ElementTree.find``
        documentation for examples and supported XPATH description.

        The most common PDS4 namespaces (those with prefix pds4, disp and sp) are registered automatically.
        If match contains other namespace prefixes (specified via colon prior to element, e.g. disp:element)
        then you must pass in `namespaces` parameter specifying the URI for each prefix.

        Parameters
        ----------
        match : str
            XPATH search string.
        namespaces : dict, optional
            Dictionary with keys corresponding to URI and values corresponding to prefix for namespace.
        unmodified : bool, optional
            Searches unmodified ``ElementTree`` root if True. Defaults to False.
        return_ET : bool, optional
            Returns an ``ElementTree`` Element instead of a Label if True. Defaults to False.

        Returns
        -------
        Label or ElementTree Element
            Matched subelement, or None if there is no match.

        """

        # Select the proper XML root to look in
        root = self.getroot(unmodified)

        # Append known namespaces if search contains them
        namespaces = self._append_known_namespaces(match, namespaces)

        # Find the matching element
        try:
            found_element = root.find(match, namespaces=namespaces)

        # Implement namespaces support for Python 2.6
        except TypeError:

            if namespaces is not None:
                match = self._add_namespaces_to_match(match, namespaces)

            found_element = root.find(match)

        # If return_ET is not used, find the other matching element
        if not return_ET and found_element is not None:

            other_element = self._find_other_element(found_element, unmodified)

            if unmodified:
                label = Label(other_element, found_element, self._namespace_map)
            else:
                label = Label(found_element, other_element, self._namespace_map)

            found_element = label

        return found_element

    def findall(self, match, namespaces=None, unmodified=False, return_ET=False):
        """ Search for all matching subelements.

        Uses the same format as ``ElementTree.findall``. See `Label` docstring or ``ElementTree.findall``
        documentation for examples and supported XPATH description.

        The most common PDS4 namespaces (those with prefix pds4, disp and sp) are registered automatically.
        If match contains other namespace prefixes (specified via colon prior to element, e.g. disp:element)
        then you must pass in `namespaces` parameter specifying the URI for each prefix.

        Parameters
        ----------
        match : str
            XPATH search string
        namespaces : dict, optional
            Dictionary with keys corresponding to URI and values corresponding to prefix for namespace.
        unmodified : bool, optional
            Searches unmodified ``ElementTree`` root if True. Defaults to False.
        return_ET : bool, optional
            Returned list contains ``ElementTree`` Elements instead of a Labels if True. Defaults to False.

        Returns
        -------
        List[Label or ElementTree Element]
            Matched subelements, or [] if there are no matches.

        """

        # Select the proper XML root to look in
        root = self.getroot(unmodified)

        # Append known namespaces if search contains them
        namespaces = self._append_known_namespaces(match, namespaces)

        # Find the matching elements
        try:
            found_elements = root.findall(match, namespaces=namespaces)

        # Implement namespaces support for Python 2.6
        except TypeError:

            if namespaces is not None:
                match = self._add_namespaces_to_match(match, namespaces)

            found_elements = root.findall(match)

        # If return_ET is not used, find the other matching elements
        if not return_ET and found_elements is not None:

            labels = []

            for element in found_elements:
                other_element = self._find_other_element(element, unmodified)

                if unmodified:
                    label = Label(other_element, element, self._namespace_map)
                else:
                    label = Label(element, other_element, self._namespace_map)

                labels.append(label)

            found_elements = labels

        return found_elements

    def to_string(self, unmodified=False, pretty_print=False):
        """ Generate a string representation of XML label.

        Parameters
        ----------
        unmodified : bool, optional
            Generates representation of unmodified ``ElementTree`` root if True. Defaults to False.
        pretty_print : bool, optional
            String representation is pretty-printed if True. Defaults to False.

        Returns
        -------
        unicode or str
            String representation of `Label`.
        """

        # Obtain root element, while also effectively deepcopying it such that any modifications made
        # inside this method (e.g. during pretty printing) do not affect the original copy.
        root = ET.fromstring(ET.tostring(self.getroot(unmodified), encoding='UTF-8'))

        # Adapted from effbot.org/zone/element-lib.htm#prettyprint, modified to properly dedent multi-line
        # values with different indents than used by this method (e.g., each line is indented by 4 spaces
        # but this method uses 2), and to preserve a single extra newline anywhere one or more are present.
        def pretty_format_xml(elem, level=0):

            i = level * '  '

            # Preserve single extra newline at the end of text (e.g. <Field>\n\n<offset>var</offset>...)
            i_text = '\n' + i
            if elem.text:

                if elem.text.count('\n') > 1:
                    i_text = ('\n' * 2) + i

            # Preserve single extra newline at the end of an element (e.g. <Field>var</Field>\n\n<offset>...)
            i_tail = '\n' + i
            if elem.tail:
                if elem.tail.count('\n') > 1:
                    i_tail = ('\n' * 2) + i

            # Remove extra space on each newline in multi-line text values
            if elem.text and elem.text.strip() and (elem.text.count('\n') > 0 or elem.text.count('\r') > 0):
                lines = elem.text.splitlines(True)

                # If each line beginning contains extra space (e.g., it used 4 spaces to align but we use 2)
                # then remove that extra space
                for j, line in enumerate(lines):
                    if line[0:len(i)] == i:
                        lines[j] = line[len(i)-1:]

                elem.text = ''.join(lines)

                # On last line of multi-line string, if it ends in a newline and spaces, remove them
                # and replace with a newline and spaces such that the closing tag is on the same level
                # as the opening tag
                if elem.text.rstrip(' ') != elem.text.rstrip():
                    elem.text = elem.text.rstrip() + '\n' + i

            if len(elem):

                if not elem.text or not elem.text.strip():
                    elem.text = i_text + '  '
                if not elem.tail or not elem.tail.strip():
                    elem.tail = i_tail
                for elem in elem:
                    pretty_format_xml(elem, level+1)
                if not elem.tail or not elem.tail.strip():
                    elem.tail = i_text
            else:
                if level and (not elem.tail or not elem.tail.strip()):
                    elem.tail = i_tail

        # In Python 2, we converted upon label read-in any unicode strings to byte strings. This means
        # that even with encoding given, `ET.tostring` will fail on any unicode string. Therefore we
        # convert everything to unicode first.
        if six.PY2:

            for elem in root.getiterator():

                if isinstance(elem.tag, bytes):
                    elem.tag = elem.tag.decode('UTF-8')

                if isinstance(elem.text, bytes):
                    elem.text = elem.text.decode('UTF-8')

                if isinstance(elem.tail, bytes):
                    elem.tail = elem.tail.decode('UTF-8')

        # Pretty-format root node
        if pretty_print:
            pretty_format_xml(root)

        # Remove newline characters from tail of the root element
        if root.tail is not None:
            root.tail = root.tail.strip('\n\r')

        # Obtain string representation, taking care to register_namespaces()
        self._register_namespaces('register')
        string = ET.tostring(root, encoding='UTF-8').decode('UTF-8')
        self._register_namespaces('unregister')

        # Remove UTF-8 processing instruction from first line
        string_list = string.splitlines(True)
        string = ''.join(string_list[1:])

        # If label is not pretty printed, then we fix a potential issue with indented labels and ET
        # (The indent for an element is in the .tail attribute of the previous element, but for the
        # root element there is no element above it with a tail. We check what is the indent for the
        # closing tag of the root element and manually insert it for the opening tag.)
        if not pretty_print:
            last_tail = root[-1].tail

            if last_tail is not None:
                num_leading_spaces = len(last_tail) - len(last_tail.rstrip(' '))
                num_leading_tabs = len(last_tail) - len(last_tail.rstrip('\t'))
                # Append string together carefully because if we used ``str.format`` it would require
                # worrying about unicode on Python 2
                string = ' ' * num_leading_spaces + '\t' * num_leading_tabs + string

        return string

    def to_dict(self, unmodified=False):
        """ Generate an `OrderedDict` representation of XML label.

        Parameters
        ----------
        unmodified : bool, optional
            Generates representation of unmodified ``ElementTree`` root if True. Defaults to False.

        Returns
        -------
        OrderedDict
            A dictionary representation of `Label`, including attributes.
        """

        root = self.getroot(unmodified)

        # Achieve equivalent of register_namespaces() via tag_modify
        tag_modify = []
        for uri, prefix in self._namespace_map.items():
            if prefix.strip():
                tag_modify.append(('{{{0}}}'.format(uri), '{0}:'.format(prefix)))
            else:
                tag_modify.append(('{{{0}}}'.format(uri), ''))

        return xml_to_dict(root, tag_modify=tuple(tag_modify))

    def _find_other_element(self, element, was_unmodified):
        """
        When using `find` or `findall`, we initially search one of the two ``ElementTree`` representations
        (`_convenient_root` or `_unmodified_root`). The purpose of this method is to find the Element
        in the other representation.

        Parameters
        ----------
        element : Label or ElementTree Element
            Element to find
        was_unmodified : bool
            True if `element` was taken from `_unmodified_root`, False otherwise.


        Returns
        -------
        ``ElementTree`` Element
            Matched element.
        """

        other_element = None

        if was_unmodified:
            root = self.getroot(unmodified=True)
            other_root = self.getroot(unmodified=False)
        else:
            root = self.getroot(unmodified=False)
            other_root = self.getroot(unmodified=True)

        # Loop over all elements in root of element to find its number
        node_number = -1

        for i, child_elem in enumerate(root.getiterator()):

            if element == child_elem:
                node_number = i
                break

        # Loop over all elements of other_root to find the matching node_number
        for i, child_elem in enumerate(other_root.getiterator()):

            if i == node_number:
                other_element = child_elem
                break

        return other_element

    def _register_namespaces(self, action='register'):
        """
        Registers or unregisters namespaces via ET's ``register_namespace``. The unregister functionality is
        provided because register_namespace is global, affecting any other ET usage. Register and unregister
        allow anything in-between them to effectively have a local namespace register.

        Parameters
        ----------
        action : str
            Either 'register' or 'unregister', specifying what action to take.

        Returns
        -------
        None
        """

        prefixes = ET._namespace_map.values()
        uris = ET._namespace_map.keys()

        for uri, prefix in self._namespace_map.items():

            # Register namespaces
            if action == 'register':

                # Check if namespace prefix or URI already exists, then do not register the namespace if so.
                # This allows local namespaces that have the same prefix as a global prefix but a
                # different URI, which is valid in XML, to stay unique)
                if (prefix in prefixes) or (uri in uris):
                    continue

                # Register the namespace
                ET._namespace_map[uri] = prefix

            # Unregister the namespaces
            else:

                if (prefix in prefixes) and (uri in ET._namespace_map.keys()):
                    del ET._namespace_map[uri]

    @classmethod
    def _append_known_namespaces(cls, match, namespaces):
        """
        Appends known namespaces (i.e., those in the constant PDS4_NAMESPACES) to `namespaces` if `match`
        contains prefixes (signified by the colon). In case of duplicates, `namespaces` contents
        overwrite those of PDS4_NAMESPACES.

        Parameters
        ----------
        match : str
            XPATH search string.
        namespaces : dict
            Dictionary with keys corresponding to URI and values corresponding to prefix for namespace.

        Returns
        -------
        dict
            New namespaces dictionary containing previous `namespaces` as well as PDS4_NAMESPACES.

        """

        if (':' in match) and (namespaces is None):
            namespaces = PDS4_NAMESPACES

        elif ':' in match:
            namespaces = dict(chain(PDS4_NAMESPACES.viewitems(), namespaces.viewitems()))

        return namespaces

    @classmethod
    def _add_namespaces_to_match(cls, match, namespaces):
        """
        Python 2.6 does not support the namespaces parameter for ``ElementTree``'s `find` and `findall`,
        which are used in the implementation of `Label`'s `find` and `findall`. To add this support,
        in `match` we replace the prefix with the URI for that prefix contained in brackets.
        This is also how ``ElementPath`` works in Python 2.7 and above, from which this code is adapted.

        Parameters
        ----------
        match : str
            XPATH search string.
        namespaces : dict
            Dictionary with keys corresponding to URI and values corresponding to prefix for namespace.

        Returns
        -------
        str
            A new XPATH search string, with prefix replaced by {URI}.

        Examples
        --------
        >>> match = './/disp:Display_Settings'
        >>> namespaces = {'disp': 'http://pds.nasa.gov/pds4/disp/v1', 'sp': 'http://pds.nasa.gov/pds4/sp/v1'}

        >>> match = self._add_namespaces_to_match(match, namespaces)
        >>> print match
        .//{http://pds.nasa.gov/pds4/disp/v1}Display_Settings
        """

        xpath_tokenizer_re = re.compile("("
                                        "'[^']*'|\"[^\"]*\"|"
                                        "::|"
                                        "//?|"
                                        "\.\.|"
                                        "\(\)|"
                                        "[/.*:\[\]\(\)@=])|"
                                        "((?:\{[^}]+\})?[^/\[\]\(\)@=\s]+)|"
                                        "\s+")
        modified_match = ''

        for token in xpath_tokenizer_re.findall(match):
            tag = token[1]

            if tag and tag[0] != '{' and ':' in tag:

                try:
                    prefix, uri = tag.split(":", 1)
                    if not namespaces:
                        raise KeyError

                    modified_match += '{0}{{{1}}}{2}'.format(token[0], namespaces[prefix], uri)
                except KeyError:
                    raise SyntaxError('prefix {0} not found in prefix map'.format(prefix))
            else:
                modified_match += '{0}{1}'.format(*token)

        return modified_match


def get_display_settings_for_lid(local_identifier, label):
    """ Search a PDS4 label for Display_Settings of a data structure with local_identifier.

    Parameters
    ----------
    local_identifier : str
        The local identifier of the data structure to which the display settings belong.
    label : Label or ElementTree Element
        Label for a PDS4 product with-in which to look for the display settings.

    Returns
    -------
    Label or ElementTree Element
        Found Display_Settings section with same return type as `label` or None if not found.
    """

    matching_display = None

    # Find all the Display Settings classes in the label
    displays = label.findall('.//disp:Display_Settings')
    if not displays:
        return None

    # Find the particular Display Settings for this LID
    for display in displays:

        # Look in both PDS and DISP namespace due to standards changes in the display dictionary
        lid_disp = find_xml_element(display, './/disp:local_identifier_reference')
        lid_pds = find_xml_element(display, './/local_identifier_reference')

        if local_identifier in (lid_disp, lid_pds):
            matching_display = display
            break

    return matching_display


def get_spectral_characteristics_for_lid(local_identifier, label):
    """ Search a PDS4 label for Spectral_Characteristics of a data structure with local_identifier.

    Parameters
    ----------
    local_identifier : str
        The local identifier of the data structure to which the spectral characteristics belong.
    label : Label or ElementTree Element
        Label for a PDS4 product with-in which to look for the spectral characteristics.

    Returns
    -------
    Label or ElementTree Element
        Found Spectral_Characteristics section with same return type as `label`, or None if not found.
    """

    matching_spectral = None

    # Find all the Spectral Characteristics classes in the label
    spectra = label.findall('.//sp:Spectral_Characteristics')
    if not spectra:
        return None

    # Find the particular Spectral Characteristics for this LID
    for spectral in spectra:

        # There may be multiple local internal references for each spectral data object sharing these
        # characteristics. Also look in both PDS and SP namespace due to standards changes in the spectral
        # dictionary
        references = spectral.findall('sp:Local_Internal_Reference') + \
                     spectral.findall('Local_Internal_Reference')

        for reference in references:

            # Look in both PDS and DISP namespace due to standards changes in the display dictionary
            lid_sp = find_xml_element(reference, 'sp:local_identifier_reference')
            lid_pds = find_xml_element(reference, 'local_identifier_reference')

            if local_identifier in (lid_sp, lid_pds):
                matching_spectral = spectral
                break

    return matching_spectral


def get_mission_area(label):
    """ Search a PDS4 label for a Mission_Area.

    Parameters
    ----------
    label : Label or ElementTree Element
        Label for a PDS4 product with-in which to look for a mission area.

    Returns
    -------
    Label or ElementTree Element
        Found Mission_Area section with same return type as `label`, or None if not found.
    """

    return label.find('.//Mission_Area')


def get_discipline_area(label):
    """ Search a PDS4 label for a Discipline_Area.

    Parameters
    ----------
    label : Label or ElementTree Element
        Label for a PDS4 product with-in which to look for a discipline area.

    Returns
    -------
    Label or ElementTree Element
        Found Discipline_Area section with same return type as `label`, or None if not found.
    """
    return label.find('.//Discipline_Area')
