from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math
import functools

from ..extern import six

# Safe import of OrderedDict
try:
    from collections import OrderedDict
except ImportError:
    from ..extern.ordered_dict import OrderedDict

# Safe import of numpy (not required)
try:
    import numpy as np
except ImportError:
    np = None

#################################


def cast_int_float_string(value):
    """ Cast given string value, if possible, to an int, float or returns unchanged.

    Parameters
    ----------
    value : str
        Value to try casting to an int and float.

    Returns
    -------
    int, float or str
        Cast value.
    """

    try:
        return int(value)

    except ValueError:

        try:
            return float(value)

        except ValueError:
            return value


def is_array_like(value):
    """
    Array-like values are defined as those that implement __len__ (such as ``list``, ``tuple``,
    ``array.array``, ``np.ndarray``, etc) but are not ``str``, ``unicode`` or ``bytes``.

    Parameters
    ----------
    value
        Any kind of value.

    Returns
    -------
    bool
        True if `value` is array-like, false otherwise.
    """

    # Checks if value implements __len__, and ensures its not a string (six is used because
    # str, unicode, and bytes can all represent strings, depending on the Python version)
    if hasattr(value, '__len__') and (not isinstance(value, (six.binary_type, six.text_type))):
        return True

    return False


def finite_min_max(array_like):
    """ Obtain finite (non-NaN, non-Inf) minimum and maximum of an array.

    Parameters
    ----------
    array_like : array_like
        A numeric array of some kind, possibly containing NaN or Inf values.

    Returns
    -------
    min, max
        Tuple of the finite minimum and maximum.
    """

    # Determine finite min and max efficiently via NumPy
    if np is not None:

        if not isinstance(array_like, np.ndarray):
            array_like = np.asarray(array_like)

        finite_values = array_like[np.isfinite(array_like)]

        return finite_values.min(), finite_values.max()

    # Determine finite min and max if NumPy is unavailable
    else:

        finite_values = [value for value in array_like
                         if (not math.isinf(value)) and (not math.isnan(value))]

        return min(finite_values), max(finite_values)


def find_xml_element(xml, key):
    """
    Convenience function to get the text representation of an XML element, similar
    to dict.get(key, None) but for `Label`s or ``ElementTree``.

    Parameters
    ----------
    xml : Label or ElementTree Element
        XML representation to find an Element in.
    key : str

    Returns
    -------
    Label or ElementTree Element
        Matched subelement's text, with same type as input `xml`. Or None if no match was found.
    """

    find_result = xml.find(key)

    if xml.find(key) is not None:
        return find_result.text

    return None


def xml_to_dict(xml_element, cast_values=False, skip_attributes=False, tag_modify=()):
    """ Transforms XML to an ``OrderedDict``.

    Takes an XML ``ElementTree`` Element or a `Label` and creates an equivalent ``OrderedDict``.
    Keys of the dictionary represent tag names and values represent the text values of the elements.
    In case of a (sub)element having child elements, values will be another ``OrderedDict``, inside
    which the text of the element has key '_text'. In case of (sub)elements having child elements with
    the same key, the value for the key will be a ``list``. In case of (sub)elements with attributes,
    the value will be an ``OrderedDict``, inside which the key for each attribute starts with '@' and
    the text of the element has key '_text'. For text elements, the text value is not preserved (and
    a '_text' key is not created) if it contains only whitespace (including spaces, tabs and newlines);
    otherwise whitespaces are preserved.

    Preserves order of elements in most cases. The exception is when an element has 2 or more sets
    of children, where each set has the same key names (i.e., there are at least 4 children, and 2 of
    those children have one key, and 2 have another key) and the order of the children with the
    non-matching keys is intertwined, in such a case the order of the intertwined keys will not be preserved.

    Adapted from http://stackoverflow.com/a/10076823.

    Parameters
    ----------
    xml_element : ``ElementTree`` Element or Label
        XML representation which will be turned into a dictionary.
    cast_values : bool, optional
        If True, float and int values will be cast as such inside output dictionary. Defaults to False.
    skip_attributes : bool, optional
        If True, skips adding attributes from XML. Defaults to False.
    tag_modify : tuple, optional
        If given, then a 2-valued tuple, or a tuple of 2-valued tuples; any match in element tag names
        and/or attributes names for each tag_modify[0] is replaced with tag_modify[1]. Empty by default.

    Returns
    -------
    OrderedDict
        Dictionary representation of the XML input.
    """

    # Modify tags if requested
    element_tag = xml_element.tag

    if tag_modify:

        if not is_array_like(tag_modify[0]):
            tag_modify = (tag_modify, )

        for tag in tag_modify:
            element_tag = element_tag.replace(tag[0], tag[1])

    d = {element_tag: OrderedDict() if xml_element.attrib else None}
    children = list(xml_element)

    # Add children
    if children:

        dd = OrderedDict()

        xml_to_dict_func = functools.partial(xml_to_dict, cast_values=cast_values,
                                             skip_attributes=skip_attributes, tag_modify=tag_modify)

        for dc in map(xml_to_dict_func, children):
            for k, v in six.iteritems(dc):
                try:
                    dd[k].append(v)
                except KeyError:
                    dd[k] = [v]

        ddd = OrderedDict()

        for k, v in six.iteritems(dd):
            if len(v) == 1:
                ddd[k] = v[0]
            else:
                ddd[k] = v

        d = {element_tag: ddd}

    has_attribs = xml_element.attrib and not skip_attributes

    # Add attributes
    if has_attribs:

        attrib = OrderedDict()
        if tag_modify:
            for k, v in six.iteritems(xml_element.attrib):

                new_k = k
                for tag in tag_modify:
                    new_k = new_k.replace(tag[0], tag[1])

                attrib[new_k] = v
        else:
            attrib = xml_element.attrib

        d[element_tag].update(('@' + k, v) for k, v in six.iteritems(attrib))

    # Add text elements
    text = xml_element.text
    if (text is not None) and (text.strip()):

        if cast_values:
            text = cast_int_float_string(text)

        if children or has_attribs:
            if text:
                d[element_tag]['_text'] = text

        else:
            d[element_tag] = text

    return d
