from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import sys
import logging

from ..extern import six

# Minimum logging levels for loud and quiet operation
_loud = logging.DEBUG
_quiet = logging.ERROR


def logger_init():
    """ Initializes the logger and its handlers.

    Returns
    -------
    logger
        The global logger for all pds4 tools.
    """

    logging.setLoggerClass(PDS4Logger)
    logger = logging.getLogger('PDS4Logger')
    logger.setLevel(_loud)

    # If this is a new logger then initialize its config
    if not logger.handlers:

        # Create the stdout handler. This handler outputs to stdout and to warning and errors message boxes
        # in the viewer and can be silenced by user (via use of quiet or --quiet).
        stdout_handler = PDS4StreamHandler('stdout_handler')

        # Create the log handler. This handler does not output to stdout or to screen and should not
        # be silenced.
        log_handler = PDS4SilentHandler('log_handler')

        # Create the formatter and add it to the handlers
        formatter = PDS4Formatter()
        stdout_handler.setFormatter(formatter)
        log_handler.setFormatter(formatter)

        # Add handlers to logger.
        logger.addHandler(stdout_handler)
        logger.addHandler(log_handler)

    return logger


class PDS4Logger(logging.Logger):
    """ Customer Logger that supports getting handlers by name and has quiet() and loud() methods. """

    def get_handler(self, handler_name):

        for handler in self.handlers:

            if handler.name == handler_name:
                return handler

        return None

    def quiet(self, handler_name='stdout_handler'):
        self.get_handler(handler_name).setLevel(_quiet)

    def loud(self, handler_name='stdout_handler'):
        self.get_handler(handler_name).setLevel(_loud)

    def is_quiet(self, handler_name='stdout_handler'):
        return self.get_handler(handler_name).is_quiet


class PDS4StreamHandler(logging.StreamHandler):
    """ Custom StreamHandler that has a self.name and a self.is_quiet. """

    def __init__(self, name):

        # Using try due to stream parameter being renamed in Python <2.7)
        try:
            logging.StreamHandler.__init__(self, stream=sys.stdout)
        except TypeError:
            logging.StreamHandler.__init__(self, strm=sys.stdout)

        self._name = name
        self._is_quiet = False

    @property
    def name(self):
        return self._name

    @property
    def is_quiet(self):
        return self._is_quiet

    def set_level(self, level):
        self.setLevel(level)

    def setLevel(self, level):

        if level == _quiet:
            self._is_quiet = True
        else:
            self._is_quiet = False

        logging.StreamHandler.setLevel(self, level)


class PDS4SilentHandler(PDS4StreamHandler):
    """ Custom StreamHandler that saves emitted records to self.records.

    Able to print out previously emitted records via `to_string`. """

    def __init__(self, name):

        PDS4StreamHandler.__init__(self, name)

        self.records = []
        self._recording_start = None

    def emit(self, record):
        """
        Parameters
        ----------
        record : logger.LogRecord

        Returns
        -------
        None
        """
        self.records.append(record)

    def begin_recording(self):
        self._recording_start = len(self.records)

    def get_recording(self, reset=True):

        record_start = self._recording_start

        if reset:
            self.begin_recording()

        return self.to_string(start_i=record_start)

    def to_string(self, start_i=0, end_i=None):
        """
        Parameters
        ----------
        start_i : int, optional
            Index of first record to include. Defaults to 0 (include records from the beginning).
        end_i : int, optional
            Index of last record to include. Defaults to None (include records until the end).

        Returns
        -------
        str
            A string containing the messages in the records that were previously emitted.
        """

        formatted_records = [self.format(record) for record in self.records[start_i:end_i]]

        return '\n'.join(formatted_records)


class PDS4Formatter(logging.Formatter):
    """ Custom formatter that varies format according to log level. """

    def format(self, record):
        """
        Parameters
        ----------
        record : logger.LogRecord
            The record to format.

        Returns
        -------
        str
            The formatted record string.

        """
        formatter = logging.Formatter('%(message)s')
        formatted_value = logging.Formatter.format(formatter, record)

        if record.levelno != logging.INFO:
            formatted_value = record.levelname.capitalize() + ': ' + formatted_value

        # On Python 2, we decode ``unicode`` to ``str`` (the emitted message retains the str type that it
        # used on being emitted), such that mixing the two in `PDS4SilentHandler.to_string` does not
        # cause an error.
        if six.PY2 and isinstance(formatted_value, unicode):
            formatted_value = formatted_value.encode('utf-8')

        return formatted_value
