from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math
import os
import weakref
import platform
from fractions import Fraction

import numpy as np
import matplotlib as mpl
from matplotlib.backends.backend_tkagg import NavigationToolbar2TkAgg
from matplotlib.figure import Figure

from .core import Window, DataViewWindow
from .mpl import FigureCanvas

from ..utils.helpers import finite_min_max
from ..utils.exceptions import PDS4StandardsException
from ..utils.logging import logger_init

from ..extern.zscale import zscale
from ..extern import six
from ..extern.six.moves.tkinter import (Menu, Frame, Scrollbar, Label, Entry, Scale, Button,
                                          Radiobutton, StringVar, BooleanVar, DoubleVar, IntVar)
from ..extern.six.moves.tkinter_tkfiledialog import asksaveasfilename

# Safe import of OrderedDict
try:
    from collections import OrderedDict
except ImportError:
    from ..extern.ordered_dict import OrderedDict

# Safe import of PowerNorm (available in MPL v1.4+)
try:
    from matplotlib.colors import PowerNorm

except ImportError:
    PowerNorm = None

# Initialize the logger
logger = logger_init()

#################################


class ImageViewWindow(DataViewWindow):
    """ Window that displays Array_2D and Array_3D PDS4 data structures as an image.

    This window will display image data. After creating it, you should use the load_array() method to load
    the array that it needs to display.
    """

    def __init__(self, parent):

        # Create basic PDS4 Viewer
        super(ImageViewWindow, self).__init__(parent)

        # Pack the display frame, which contains the scrollbars and the scrollable canvas
        self._canvas.config(bg='white')
        self._display_frame.config(bg='white')
        self._display_frame.pack(side='left', anchor='nw', expand=1, fill='both')

        # Control variable for the `_image_display_manage` function
        self._image_display_locked = False

        # Will be set to an instance of FigureCanvas (containing the main image/slice being displayed)
        self._figure_canvas = None

        # Will be set to an instance of MPL's AxesImage
        self._image = None

        # Will be set to an instance of Colorbar
        self._colorbar = None

        # Will be set to an instance of MPL's toolbar for TK
        self._toolbar = None

        # Will be set to the data for the current slice/image being shown
        self._slice_data = None

        # Contains sub-widgets of the header
        self._header_widgets = {'x': None, 'y': None, 'value': None, 'frame': None}

        # Menu option variables. These are TKinter type wrappers around standard Python variables. The
        # advantage is that you can use trace(func) on them, to automatically call func whenever one of
        # these variables is changed
        update_colorbar = lambda *args: self._update_colorbar(full_redraw=True)

        menu_options = [
            {'name': 'show_colorbar',   'type': BooleanVar(), 'default': True,         'trace': self.toggle_colorbar},
            {'name': 'colorbar_orient', 'type': StringVar(),  'default': 'horizontal', 'trace': update_colorbar},
            {'name': 'colormap',        'type': StringVar(),  'default': 'gray',       'trace': self._update_colormap},
            {'name': 'invert_colormap', 'type': BooleanVar(), 'default': False,        'trace': self._update_colormap},
            {'name': 'show_border',     'type': BooleanVar(), 'default': False,        'trace': self._update_tick_labels},
            {'name': 'show_axis_ticks', 'type': StringVar(),  'default': 'none',       'trace': self._update_tick_labels},
            {'name': 'scale',           'type': StringVar(),  'default': 'linear',     'trace': self._update_scale},
            {'name': 'scale_limits',    'type': StringVar(),  'default': 'min max',    'trace': self._update_scale},
            {'name': 'zoom_level',      'type': DoubleVar(),  'default': 1,            'trace': self._update_zoom},
            {'name': 'invert_axis',     'type': StringVar(),  'default': 'none',       'trace': self._update_invert_axis},
            {'name': 'rotation_angle',  'type': DoubleVar(),  'default': 0,            'trace': self._update_rotation},
            {'name': 'orientation',     'type': StringVar(),  'default': 'display',    'trace': self._update_orientation},
            {'name': 'pan',             'type': BooleanVar(), 'default': False,        'trace': self._pan},
            {'name': 'axis_limits_to_rectangle', 'type': BooleanVar(), 'default': False, 'trace': self._axis_limits_to_rectangle}
        ]

        for option in menu_options:

            var = option['type']
            self._menu_options[option['name']] = var

            self._add_trace(var, 'w', option['trace'], option['default'])

    @property
    def settings(self):

        settings = super(ImageViewWindow, self).settings
        settings['axes'] = settings['axes'].copy()

        return settings

    # Loads the image array structure into this window and displays it for the user
    def load_array(self, array_structure):

        # Set a title for the window
        self._set_title("{0} - Image '{1}'".format(self._widget.title(), array_structure.id))

        # Set necessary instance variables for this DataViewWindow
        self.data = array_structure.data
        self.meta_data = array_structure.meta_data
        self.structure = array_structure
        self._settings = {'axes': _AxesProperties(), 'selected_axis': 0,
                          'is_rgb': False, 'rgb_bands': (None, None, None),
                          'dpi': 80., 'pixel_init_dimensions': (0, 0), 'pixel_margins': (0, 0)}

        # Create the header
        self._draw_header()

        # Add vertical scrollbar for the image
        self._vert_scrollbar = Scrollbar(self._display_frame, orient='vertical', command=self._scrollable_canvas.yview)
        self._scrollable_canvas.config(yscrollcommand=self._vert_scrollbar.set)
        self._vert_scrollbar.pack(side='right', fill='y')

        # Add horizontal scrollbar for the image
        self._horz_scrollbar = Scrollbar(self._display_frame, orient='horizontal', command=self._scrollable_canvas.xview)
        self._scrollable_canvas.config(xscrollcommand=self._horz_scrollbar.set)
        self._horz_scrollbar.pack(side='bottom', fill='x')

        # Pack the static canvas, which contains the scrollable canvas and the colorbar
        self._static_canvas.config(background='white')
        self._static_canvas.pack(side='left', anchor='nw', expand=1, fill='both')

        # Place the scrollable canvas, which contains the image itself
        self._scrollable_canvas.config(background='white')
        self._scrollable_canvas.place(relx=0.5, rely=0.5, anchor='center')

        # Update idletasks such that the window takes on its real size
        self._widget.update_idletasks()

        # Set horizontal and vertical axes for each slice
        self.set_slice_axes()

        # Special settings initialization for RGB images
        color_axis = self._settings['axes'].find('type', 'color')
        if color_axis is not None:

            self._settings['is_rgb'] = True

            # Set RGB bands for color images
            self.set_rgb_bands()

            # Disable default showing of colorbar for RGB images (since works only for monochrome images)
            self._menu_options['show_colorbar'].set(False)

            # Set colormap to RGB by default
            self.set_colormap('RGB')

        # Draw the slice/image on the screen
        self._draw_slice()

        # Add notify event for window resizing
        self._display_frame.bind('<Configure>', self._window_resize)

        # Add notify event for scroll wheel (used to zoom via scroll wheel)
        self._bind_scroll_event(self._mousewheel_scroll)

        # Add notify event for mouse pointer exiting _scrollable_canvas (used to display pixel values
        # under the mouse pointer)
        self._figure_canvas.tk_widget.bind('<Leave>', self._update_mouse_pixel_value)

        # Add notify event for mouse motion (used to display pixel values under pointer)
        self._figure_canvas.mpl_canvas.mpl_connect('motion_notify_event', self._update_mouse_pixel_value)

        self._add_menu()
        self._data_open = True

    # Sets colormap on currently displayed image
    def set_colormap(self, colormap):
        self._menu_options['colormap'].set(colormap)

    # Sets the RGB bands for the color axis, if image has one (e.g. if color axis has 5 bands, this
    # determines which 3 bands will be used for the RGB image). Each color parameter must be set an integer
    # indice of the color axis (1-based indexing, as per the display dictionary) or None. None will default
    # to display settings values.
    def set_rgb_bands(self, red=None, green=None, blue=None):

        # Raise error if image is not RGB (i.e., has no color axis)
        if not self._settings['is_rgb']:
            raise TypeError('Unable to set RGB bands for an image without a color axis.')

        display_settings = self.meta_data.display_settings
        color_settings = display_settings['Color_Display_Settings']

        if red is None:
            red = color_settings['red_channel_band']

        if green is None:
            green = color_settings['green_channel_band']

        if blue is None:
            blue = color_settings['blue_channel_band']

        # Attempt to correct zero-based (or less) indexing
        min_index = min(red, green, blue)
        if min_index < 0:
            adjust_index = 1 - min_index
            red += adjust_index
            green += adjust_index
            blue += adjust_index
            self._issue_warning('Found non-compliant (non-1) based indexing in Color Settings. '
                                'Displayed image may not be correct.', show=False)

        # Save the newly input values
        self._settings['rgb_bands'] = (red, green, blue)

        # Reload the slice to reflect the changes
        if self._data_open:
            self._draw_slice()

    # Controls scaling for the currently displayed slice. scale can have values linear|log|squared|square root,
    # and limits can have values min max|zscale, or a string with format 'min,max'. To keep existing values
    # use None.
    def set_scale(self, scale=None, limits=None):

        if scale is not None:
            self._menu_options['scale'].set(scale)

        if limits is not None:
            self._menu_options['scale_limits'].set(limits)

    # Zooms in on the initial image dimensions by zoom_level (e.g. zoom_level of 2 will double the dimensions
    # of the original image)
    def set_zoom(self, zoom_level):
        self._menu_options['zoom_level'].set(zoom_level)

    # Rotates currently displayed slice to rotation_angle from initial orientation, in degrees. Note that
    # rotation angle must be evenly divisible by 90.
    def set_rotation(self, rotation_angle):
        self._menu_options['rotation_angle'].set(rotation_angle)

    # Sets the default orientation, from which the rotation angle and axis inversions begin. Valid options
    # are display|storage, indicating either according to display dictionary or according to the storage
    # order
    def set_orientation(self, orientation):
        self._menu_options['orientation'].set(orientation)

    # Controls axes display for current image. show_axis and invert_axis can have values x|y|both|none.
    # show_border is a boolean that controls whether a border is displayed around image. To keep existing
    # values use None.
    def set_axes_display(self, show_axis=None, invert_axis=None, show_border=None):

        if show_axis is not None:
            self._menu_options['show_axis_ticks'].set(show_axis)

        if invert_axis is not None:
            self._menu_options['invert_axis'].set(invert_axis)

        if show_border is not None:
            self._menu_options['show_border'].set(show_border)

    # Zooms the image, via resampling, to fit the current window
    def zoom_to_fit(self):

        fig_size = self._settings['pixel_init_dimensions']
        margins = self._settings['pixel_margins']

        # Retrieve width and height of static_canvas (the width and height that we need to zoom to fit)
        self._widget.update_idletasks()
        width = self._static_canvas.winfo_width()
        height = self._static_canvas.winfo_height()

        # Adjust maximum zoom level to remove colorbar size (since the colorbar takes up some space we need
        # to fit in)
        if self._colorbar is not None:

            if self._colorbar.orientation == 'vertical':
                width -= self._colorbar.dimensions[0]

            else:
                height -= self._colorbar.dimensions[1]

        # Determine the maximum zoom level possible before image will not fit in frame
        extra_space_width = margins[0] * 2
        extra_space_height = margins[1] * 2

        fig_size_div = ((width - extra_space_width) / fig_size[0],
                        (height - extra_space_height) / fig_size[1])

        zoom_level = min(fig_size_div)
        self._zoom(zoom_level=zoom_level)

    # Sets the horizontal and vertical axes to be used for each slice. The values must be the integer
    # sequence_number of the Axis_Array, or None, which defaults to the display settings values.
    def set_slice_axes(self, horizontal_axis=None, vertical_axis=None):

        meta_data = self.meta_data
        display_settings = meta_data.display_settings
        invalid_display_settings = (not display_settings) or (not display_settings.valid)
        axis_arrays = meta_data.get_axis_arrays(sort=True)

        time_axis = None
        color_axis = None

        # In case of non-existent or invalid Display Settings, default to ([frame], vertical, horizontal)
        # order for axes in 2D and 3D if necessary. For 4D+, raise error unless sequence numbers are given
        if invalid_display_settings and (horizontal_axis is None or vertical_axis is None):

            if len(axis_arrays) > 3:
                raise PDS4StandardsException('No valid/supported Display Settings found. Unable to show '
                                             'images with more than 3 axes.')

            self._issue_warning('No valid/supported Display Settings found. Displayed image may '
                                'not be correct.', show=True)

            self._menu_options['orientation'].set('storage')

            three_d_adjust = 0
            if len(axis_arrays) == 3:
                three_d_adjust = 1

            if horizontal_axis is None:
                horizontal_axis = 2 + three_d_adjust

            if vertical_axis is None:
                vertical_axis = 1 + three_d_adjust

        # Determine horizontal and vertical axes from Display Dictionary if necessary
        elif display_settings.valid:

            # Find names of all present axes in Display Dictionary
            display_direction = display_settings['Display_Direction']
            color_settings = display_settings.get('Color_Display_Settings')
            movie_settings = display_settings.get('Movie_Display_Settings')

            if horizontal_axis is None:
                horizontal_name = display_direction['horizontal_display_axis']
                horizontal_axis = meta_data.get_axis_array(axis_name=horizontal_name)['sequence_number']

            if vertical_axis is None:
                vertical_name = display_direction['vertical_display_axis']
                vertical_axis = meta_data.get_axis_array(axis_name=vertical_name)['sequence_number']

            if color_settings is not None:
                color_name = color_settings['color_display_axis']
                color_axis = meta_data.get_axis_array(axis_name=color_name)['sequence_number']

            if movie_settings is not None:
                time_name = movie_settings['time_display_axis']
                time_axis = meta_data.get_axis_array(axis_name=time_name)['sequence_number']

        type_to_number = {
            'horizontal': horizontal_axis,
            'vertical': vertical_axis,
            'color': color_axis,
            'time': time_axis
        }

        # Add axes to the settings['axes'] helper object (_AxesProperties). Consists of essentially a list
        # with a dictionary for each axis present in the array, indicating the name, type, sequence number,
        # selected slice and length of the axis
        axes = _AxesProperties()

        for axis in axis_arrays:

            name = axis['axis_name']
            length = axis['elements']
            sequence = axis['sequence_number']

            type_ = None
            slice_ = 0

            for type, seq_number in six.iteritems(type_to_number):

                if seq_number == sequence:
                    type_ = type
                    slice_ = slice(None)

            axes.add_axis(name, type_, sequence, slice_, length)

        # Ensure that valid axes were found
        error = "Unable to set slice axes, an Axis_Array with sequence_number '{0}' does not exist."

        if axes.find('type', 'horizontal') is None:
            raise IndexError(error.format(horizontal_axis))

        elif axes.find('type', 'vertical') is None:
            raise IndexError(error.format(vertical_axis))

        # Save the new slice axes
        self._settings['axes'] = axes

        # Reload the slice to reflect the changes
        if self._data_open:
            self._draw_slice()

    # Selects a new slice to display and updates the currently displayed slice/image. To adjust the shown
    # slice for a 3D image, having axes (x, y, z), you need to specify both which axis (for example x, which
    # could correspond to time) and which slice of x (for example 2) you want to display. The axis variable
    # is an integer specifying the sequence_number of the Axis_Array, in other words it specifies for which
    # axis the index variable is used; None defaults to the currently selected axis. The index variable is
    # either an integer specifying which slice will be shown for the axis, or the string 'next' or 'previous';
    # None defaults to the currently selected slice
    def select_slice(self, axis=None, index=None):

        settings = self._settings
        axes = settings['axes']

        # Skip any action if there are only two axes (horizontal and vertical)
        if len(axes) == 2:
            return

        # Select currently selected axis by default
        if axis is None:
            axis = axes[settings['selected_axis']]['sequence_number']

        # Raise error if axis sequence_number does not exist
        axis_properties = axes.find('sequence_number', axis)

        if axis_properties is None:
            raise IndexError('Invalid axis sequence_number specified: {0}'.format(axis))

        # Select the currently selected slice for the axis by default
        if index is None:
            index = axis_properties['slice']

        previous_slice = axis_properties['slice']
        num_slices = axis_properties['length'] - 1

        if isinstance(index, str):

            if index == 'next':
                new_slice = previous_slice + 1

            elif index == 'previous':
                new_slice = previous_slice - 1

            else:
                raise IndexError('Invalid slice requested: {0}'.format(index))

            # Roll over to zero if reached the end
            if new_slice > num_slices:
                new_slice = 0

            # Roll over to end if reached the beginning
            if new_slice < 0:
                new_slice = num_slices

        else:

            if (not isinstance(index, six.integer_types)) or (index > num_slices) or (index < 0):
                raise IndexError('Invalid slice selection specified: {0}'.format(index))

            new_slice = index

        # Save new selected axis
        axis_index = axes.find_index('sequence_number', axis)
        self._settings['selected_axis'] = axis_index

        # Display a new slice if needed (changing only the selected axis, while keeping the slice index
        # the same, does not require a re-display of the data since it does not change)
        if new_slice != previous_slice:

            axes.set(axis_index, 'slice', new_slice)
            self._slice_data = self._get_slice_data(update=True)

            # Freeze the image display temporarily so it is not needlessly re-drawn multiple times
            self._image_display_manage('freeze', lock=True)

            self._header_widgets['frame'].set(new_slice)
            self._image.set_data(weakref.proxy(self._slice_data))
            self._update_scale()

            # Thaw the image display
            self._image_display_manage('thaw', lock=False)

    # Toggles whether colorbar is displayed for current image
    def toggle_colorbar(self, *args):

        if self.menu_option('show_colorbar'):
            self._add_colorbar()

        else:
            self._remove_colorbar()

    # Resets the current view of the image, including all axis-limits, pan, zoom, rotations and inversions,
    # to the original view
    def reset_view(self):

        # Freeze the image display temporarily so it is not needlessly re-drawn multiple times
        self._image_display_manage('freeze', lock=True)

        self._zoom(zoom_level=1)
        self.set_axes_display(invert_axis='none')
        self.set_rotation(0)

        # Thaw the image display
        self._image_display_manage('thaw', lock=False)

    # Adds menu options used for manipulating the data display
    def _add_menu(self):

        # Adds an Export sub-menu to the File menu
        file_menu = self._menu.winfo_children()[0]
        # export_menu = Menu(file_menu)

        # file_menu.insert_cascade(0, label='Export', menu=export_menu)
        file_menu.insert_command(0, label='Save Image', command=self._save_file_box)

        file_menu.insert_separator(1)

        # Add a Frame menu
        frame_menu = Menu(self._menu, tearoff=0)
        self._menu.add_cascade(label='Frame', menu=frame_menu)

        frame_menu.add_command(label='Data Cube', command=
                               lambda: self._add_dependent_window(DataCubeWindow(self._viewer, self)))

        frame_menu.add_separator()

        frame_menu.add_command(label='Next Frame', command=lambda: self.select_slice(index='next'))
        frame_menu.add_command(label='Previous Frame', command=lambda: self.select_slice(index='previous'))

        # Add a Zoom menu
        zoom_menu = Menu(self._menu, tearoff=0)
        self._menu.add_cascade(label='Zoom', menu=zoom_menu)

        zoom_menu.add_command(label='Reset Original View', command=self.reset_view)

        zoom_menu.add_separator()

        zoom_menu.add_command(label='Zoom to Fit', command=self.zoom_to_fit)

        zoom_menu.add_separator()

        zoom_menu.add_checkbutton(label='Pan', onvalue=True, offvalue=False, variable=self._menu_options['pan'])
        zoom_menu.add_checkbutton(label='Axis-Limits to Rectangle', onvalue=True, offvalue=False,
                                  variable=self._menu_options['axis_limits_to_rectangle'])

        zoom_menu.add_separator()

        zoom_menu.add_command(label='Zoom In', command=lambda: self._zoom('in'))
        zoom_menu.add_command(label='Zoom Out', command=lambda: self._zoom('out'))

        zoom_menu.add_separator()

        zoom_levels = (1./32, 1./16, 1./8, 1./4, 1./2, 1, 2, 4, 8, 16, 32)

        for level in zoom_levels:
            label = 'Zoom {0}'.format(Fraction.from_float(level))
            zoom_menu.add_checkbutton(label=label, onvalue=level, offvalue=level,
                                      variable=self._menu_options['zoom_level'])

        zoom_menu.add_separator()

        invert_options = OrderedDict([('None', 'none'), ('Invert X', 'x'),
                                      ('Invert Y', 'y'), ('Invert XY', 'both')])
        for label, option in six.iteritems(invert_options):

            zoom_menu.add_checkbutton(label=label, onvalue=option, offvalue=option,
                                      variable=self._menu_options['invert_axis'])

        zoom_menu.add_separator()

        rotation_angles = (0, 90, 180, 270)

        for angle in rotation_angles:
            label = '{0} Degrees'.format(angle)
            zoom_menu.add_checkbutton(label=label, onvalue=angle, offvalue=angle,
                                      variable=self._menu_options['rotation_angle'])

        zoom_menu.add_separator()

        valid_display_settings = self.meta_data.display_settings and self.meta_data.display_settings.valid
        disable_display_orient = {} if valid_display_settings else {'state': 'disabled'}

        zoom_menu.add_checkbutton(label='Display Orientation', onvalue='display', offvalue='display',
                                  variable=self._menu_options['orientation'], **disable_display_orient)
        zoom_menu.add_checkbutton(label='Storage Orientation', onvalue='storage', offvalue='storage',
                                  variable=self._menu_options['orientation'])

        # Add a Scale menu
        scale_menu = Menu(self._menu, tearoff=0)
        self._menu.add_cascade(label='Scale', menu=scale_menu)

        scale_types = ('linear', 'log', 'squared', 'square root')

        for scale in scale_types:

            # PowerNorm is an MPL feature available in MPL v1.4+ only
            disable_powernorm = {}

            if (PowerNorm is None) and (scale in ('squared', 'square root')):
                disable_powernorm = {'state': 'disabled'}

            scale_menu.add_checkbutton(label=scale.title(), onvalue=scale, offvalue=scale,
                                       variable=self._menu_options['scale'], **disable_powernorm)

        scale_menu.add_separator()

        scale_limits = ('min max', 'zscale')

        for limit in scale_limits:
            scale_menu.add_checkbutton(label=limit.title(), onvalue=limit, offvalue=limit,
                                       variable=self._menu_options['scale_limits'])

        scale_menu.add_separator()

        scale_menu.add_command(label='Scale Parameters...', command=lambda: self._add_dependent_window(
                ScaleParametersWindow(self._viewer, self, self._image)))

        # Add a Color menu
        color_menu = Menu(self._menu, tearoff=0)
        self._menu.add_cascade(label='Color', menu=color_menu)

        colormaps = ('gray', 'Reds', 'Greens', 'Blues', 'hot', 'afmhot', 'gist_heat', 'cool', 'coolwarm',
                     'jet', 'rainbow', 'hsv')

        for color in colormaps:
            color_menu.add_checkbutton(label=color, onvalue=color, offvalue=color,
                                       variable=self._menu_options['colormap'])

        color_menu.add_separator()

        color_menu.add_checkbutton(label='RGB', onvalue='RGB', offvalue='RGB',
                                   variable=self._menu_options['colormap'])

        if not self._settings['is_rgb']:
            color_menu.entryconfig(color_menu.index('last'), state='disabled')

        color_menu.add_separator()

        # Add an All Colors sub-menu to the Color menu
        all_colors_menu = Menu(self._menu, tearoff=0)
        color_menu.add_cascade(label='All Colors', menu=all_colors_menu)

        all_colormaps = [m for m in mpl.cm.datad if not m.endswith("_r")]

        for color in all_colormaps:
            all_colors_menu.add_checkbutton(label=color, onvalue=color, offvalue=color,
                                            variable=self._menu_options['colormap'])

        color_menu.add_separator()

        color_menu.add_checkbutton(label='Invert Colormap', onvalue=True, offvalue=False,
                                   variable=self._menu_options['invert_colormap'])

        color_menu.add_separator()

        # Add a Colorbar sub-menu to the Color menu
        colorbar_menu = Menu(color_menu, tearoff=0)
        color_menu.add_cascade(label='Colorbar', menu=colorbar_menu)

        colorbar_menu.add_checkbutton(label='Show', onvalue=True, offvalue=False,
                                      variable=self._menu_options['show_colorbar'])

        # Add a Colorbar orientation sub-menu to the Colorbar menu
        colorbar_orient_menu = Menu(color_menu, tearoff=0)
        colorbar_menu.add_cascade(label='Orientation', menu=colorbar_orient_menu)

        orientations = ('vertical', 'horizontal')

        for orientation in orientations:
            colorbar_orient_menu.add_checkbutton(label=orientation.capitalize(), onvalue=orientation,
                                                 offvalue=orientation,
                                                 variable=self._menu_options['colorbar_orient'])

        # Add a View menu
        view_menu = self._add_view_menu()
        view_menu.add_separator()

        view_menu.add_checkbutton(label='Show Image Border', onvalue=True, offvalue=False,
                                  variable=self._menu_options['show_border'])

        # Add a Ticks sub-menu to the View Menu
        ticks_menu = Menu(view_menu, tearoff=0)
        view_menu.add_cascade(label='Ticks', menu=ticks_menu)

        show_ticks = OrderedDict([('X', 'x'), ('Y', 'y'), ('XY', 'both')])
        for label, option in six.iteritems(show_ticks):
            ticks_menu.add_checkbutton(label='Show {0} Ticks'.format(label), onvalue=option,
                                       offvalue='none', variable=self._menu_options['show_axis_ticks'])

        # Add a Colorbar sub-menu to the View menu
        view_menu.add_cascade(label='Colorbar', menu=colorbar_menu)

    # Draws the header box, which contains the name of the structure, the current frame, the pixel location
    # that the mouse pointer is over and that pixel's value
    def _draw_header(self):

        header_box = Frame(self._header, height=82, bd=1, bg=self.get_bg('gray'), relief='raised')
        header_box.pack(fill='x', expand=1)

        value_box_params = {'background': self.get_bg('gray'), 'borderwidth': 1, 'relief': 'groove'}
        line_box_params = {'background': self.get_bg('gray'), 'height': 25}

        entry_params = {'disabledbackground': self.get_bg('gray'), 'state': 'disabled',
                        'borderwidth': 0, 'highlightthickness': 0, 'disabledforeground': 'black'}

        # Add the header line containing the structure name
        structure_line_box = Frame(header_box, **line_box_params)
        structure_line_box.pack(side='top', anchor='w', pady=(4, 4))

        structure_text_box = Frame(structure_line_box, height=20, width=70, bg=self.get_bg('gray'))
        structure_text_box.pack(side='left')

        w = Label(structure_text_box, text='Structure', bg=self.get_bg('gray'))
        w.pack(side='left')

        structure_name_box = Frame(structure_line_box, height=20, width=330, **value_box_params)
        structure_name_box.pack(side='left')

        w = Label(structure_name_box, text=self.structure.id, bg=self.get_bg('gray'))
        w.pack(side='left', padx=(10, 10))

        structure_text_box.pack_propagate(False)
        structure_name_box.pack_propagate(False)

        # Add the header line containing the current frame number
        frame_line_box = Frame(header_box, **line_box_params)
        frame_line_box.pack(side='top', anchor='w', pady=(0, 4))

        frame_text_box = Frame(frame_line_box, height=20, width=70, bg=self.get_bg('gray'))
        frame_text_box.pack(side='left')

        w = Label(frame_text_box, text='Frame', bg=self.get_bg('gray'))
        w.pack(side='left')

        frame_num_box = Frame(frame_line_box, height=20, width=330, **value_box_params)
        frame_num_box.pack(side='left')

        frame_text_box.pack_propagate(False)
        frame_num_box.pack_propagate(False)

        frame_num = StringVar()
        frame_num.set('0')
        self._header_widgets['frame'] = frame_num
        e = Entry(frame_num_box, textvariable=frame_num, **entry_params)
        e.pack(side='left', padx=(10, 10))

        # Add the header line containing X, Y and pixel value
        pixel_line_box = Frame(header_box, **line_box_params)
        pixel_line_box.pack(side='top', anchor='w', pady=(0, 4))

        pixel_text_box = Frame(pixel_line_box, height=25, width=70, bg=self.get_bg('gray'))
        pixel_text_box.pack(side='left')

        w = Label(pixel_text_box, text='Pixel', bg=self.get_bg('gray'))
        w.pack(side='left')

        w = Label(pixel_text_box, text='X', bg=self.get_bg('gray'))
        w.pack(side='right', padx=(0, 5))

        pixel_text_box.pack_propagate(False)

        x_pixel_box = Frame(pixel_line_box, height=25, **value_box_params)
        x_pixel_box.pack(side='left')

        x_pixel = StringVar()
        self._header_widgets['x'] = x_pixel
        e = Entry(x_pixel_box, textvariable=x_pixel, width=8, **entry_params)
        e.pack(side='left', padx=(10, 10))

        w = Label(pixel_line_box, text='Y', bg=self.get_bg('gray'))
        w.pack(side='left', padx=(10, 5))

        y_pixel_box = Frame(pixel_line_box, height=25, **value_box_params)
        y_pixel_box.pack(side='left')

        y_pixel = StringVar()
        self._header_widgets['y'] = y_pixel
        e = Entry(y_pixel_box, textvariable=y_pixel, width=8, **entry_params)
        e.pack(side='left', padx=(10, 10))

        w = Label(pixel_line_box, text='Value', bg=self.get_bg('gray'))
        w.pack(side='left', padx=(10, 5))

        value_box = Frame(pixel_line_box, height=25, **value_box_params)
        value_box.pack(side='left')

        pixel_value = StringVar()
        self._header_widgets['value'] = pixel_value
        e = Entry(value_box, textvariable=pixel_value, width=14, **entry_params)
        e.pack(side='left', padx=(10, 10))

    # Returns the 2D (in case of grayscale) or 3D (in case of color) array containing the data for the
    # currently shown slice. If `update` is set, the returned array will be updated to reflect the currently
    # selected menu and slice options. The returned array will always have axes in the proper storage
    # order for display (vertical axis, horizontal axis, color axis).
    def _get_slice_data(self, update=False):

        # If we are not updating the array to reflect currently selected settings then we just
        # return the previously stored values.
        if not update:
            return self._slice_data

        axes = self._settings['axes']
        axes_slice = [axes[i]['slice'] for i in range(0, len(self._settings['axes']))]

        # Obtain 2D (in case of grayscale data) or 3D (in case of color data) image data for current
        # subframe selections
        array_data = self.data[axes_slice]

        # Obtain sequence numbers for vertical, horizontal and color axes
        vertical_seq = axes.find('type', 'vertical')['sequence_number']
        horizontal_seq = axes.find('type', 'horizontal')['sequence_number']
        color_seq = None

        color_axis = axes.find('type', 'color')
        if color_axis:
            color_seq = color_axis['sequence_number']

        # Our data is now 2D or 3D, and we need to obtain order of axes for this data. For example,
        # we want (2, 8, 4) to become (0, 2, 1)
        if color_seq is None:
            seq_list = (vertical_seq, horizontal_seq)
        else:
            seq_list = (vertical_seq, horizontal_seq, color_seq)

        indicies = [i for i, x in sorted(enumerate(seq_list), key=lambda tup: tup[1])]

        vertical_seq = indicies[0]
        horizontal_seq = indicies[1]
        color_seq = None if color_seq is None else indicies[2]

        # Roll the color axis to the last position if present
        color_axis_adjust = 0

        if color_axis is not None:

            array_data = np.rollaxis(array_data, color_seq, 3)
            color_axis_adjust = 1

            # Adjust sequence numbers of other axes (the ones that could be in front)
            if horizontal_seq > color_seq:
                horizontal_seq -= 1

            if vertical_seq > color_seq:
                vertical_seq -= 1

        # Roll the horizontal axis to second position
        array_data = np.rollaxis(array_data, horizontal_seq, 2 + color_axis_adjust)

        # Adjust sequence numbers of other axes (the ones that could be in front)
        if vertical_seq > horizontal_seq:
            vertical_seq -= 1

        # Roll the vertical axis to the first position
        array_data = np.rollaxis(array_data, vertical_seq, 1 + color_axis_adjust)

        # Adjust color axis in ways necessary for MPL to acceptably plot color images
        if color_axis is not None:

            # Adjust color axis to have 3-bands (in case it had had more), one for each color in RGB
            rgb_bands = self._settings['rgb_bands']
            color_indicies = [rgb_bands[0] - 1, rgb_bands[1] - 1, rgb_bands[1] - 1]
            array_data = array_data[:, :, color_indicies]

            colormap = self.menu_option('colormap')

            # For RGB colormap we show an RGB image
            if colormap == 'RGB':

                # Calculate the finite (non-NaN and non-Inf) maximum and minimum values in the array
                array_min, array_max = finite_min_max(array_data)

                # Adjust array of integers to fit in uint8 range as MPL's imshow() expects
                array_dtype = array_data.dtype
                if np.issubdtype(array_dtype, int):

                    np.divide(255 * (array_data - array_min), (array_max - array_min), out=array_data, casting='unsafe')

                    try:
                        array_data = array_data.astype('uint8', copy=False)
                    except TypeError:
                        array_data = array_data.astype('uint8')

                # Adjust array of floats to be normalized in 0-1 range as MPL's imshow() expects
                elif np.issubdtype(array_dtype, float):
                    np.divide((array_data - array_min), (array_max - array_min), out=array_data, casting='unsafe')

            # For grayscale colormaps we convert the data to monochrome
            else:
                color_axis = None
                array_data = np.dot(array_data[..., :3], [0.299, 0.587, 0.144])

        # Rotate slice if requested
        rotation_angle = self.menu_option('rotation_angle')
        array_data = np.rot90(array_data, int(rotation_angle // 90))

        return array_data

    # Initializes or updates the MPL Figure and draws slice_data on it. This must be called any time the
    # dimensions of slice_data change, to reflect the changes in figure dimensions. load_array() must have
    # been called for the array this slice_data came from at least once prior to this method being called.
    def _draw_slice(self):

        # Freeze the image display temporarily so it is not needlessly re-drawn multiple times
        self._image_display_manage('freeze', lock=True)

        # Get the new slice data to show
        self._slice_data = self._get_slice_data(update=True)

        # Determine image display properties
        dpi = self._settings['dpi']
        x_pixels, y_pixels = self._slice_data.shape[1], self._slice_data.shape[0]
        fig_size = x_pixels / dpi, y_pixels / dpi

        # Store initial figure size
        self._settings['pixel_init_dimensions'] = (x_pixels, y_pixels)

        # Create the figure if it does not exist
        if not self._data_open:
            figure = Figure(figsize=fig_size, dpi=dpi, facecolor=self.get_bg())

            # Create the FigureCanvas, which is a wrapper for FigureCanvasTkAgg()
            self._figure_canvas = FigureCanvas(figure, master=self._scrollable_canvas)
            self._figure_canvas.tk_widget.config(background='white')
            self._scrollable_canvas.create_window(0, 0, window=self._figure_canvas.tk_widget, anchor='nw')

            # Create the toolbar (hidden, used for its pan and axis-limits to rectangle options)
            self._toolbar = NavigationToolbar2TkAgg(self._figure_canvas.mpl_canvas, self._display_frame)
            self._toolbar.update()
            self._toolbar.pack_forget()

        # Update figure dimensions if it already exists
        else:
            self._remove_colorbar()
            self._figure_canvas.clear()
            self._figure_canvas.set_dimensions(fig_size, type='inches')

        # Create an axis for the new image
        bbox = self._get_bbox_with_margins((x_pixels, y_pixels))
        ax = self._figure_canvas.mpl_figure.add_axes(bbox, axisbelow=True, frame_on=False, clip_on=False)

        # Draw the image
        self._image = ax.imshow(weakref.proxy(self._slice_data), origin='upper', interpolation='none',
                                norm=mpl.colors.Normalize(), aspect='equal')

        # Disable X and Y tick labels initially
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)

        # Add colorbar to image
        if self._colorbar is None:
            self._add_colorbar()

        # Update scale, axis inversion, axis ticks, and zoom to match current menu_options
        self._update_scale()
        self._update_invert_axis()
        self._update_tick_labels()
        self._update_zoom()

        # Thaw the image display since it was frozen above
        self._image_display_manage('thaw', lock=False)

    # Adds a colorbar to the GUI if menu_options are set to show one
    def _add_colorbar(self):

        # Skip adding a colorbar if none is suppose to be shown
        if not self.menu_option('show_colorbar'):
            return

        # Add new colorbar
        orientation = self.menu_option('colorbar_orient')
        self._colorbar = Colorbar(self._static_canvas, self._image, orientation)

        # Creating a new scroll event binding because calling FigureCanvasTkAgg() in Colorbar() rebinds it
        self._bind_scroll_event(self._mousewheel_scroll)

        # The scrollable canvas dimensions change due to colorbar taking up space and need updating
        self._update_scrollable_canvas_dimensions()

    # Removes the colorbar from the GUI if one is being shown
    def _remove_colorbar(self):

        # Skip removing the colorbar if none exists
        if self._colorbar is None:
            return

        # Remove the callback to destroy the colorbar on window close
        self._remove_callbacks('close', self._colorbar.destroy, ignore_args=True)

        self._colorbar.destroy()
        self._colorbar = None

        # The scrollable canvas dimensions change due to colorbar taking up space and need updating
        self._update_scrollable_canvas_dimensions()

    # Activates the Pan image option
    def _pan(self, *args):

        # Disable axis-limits to rectangle if it is enabled
        limits_option = self._menu_options['axis_limits_to_rectangle']
        pan_option = self._menu_options['pan']

        if pan_option.get() and limits_option.get():
            limits_option.set(False)

        # Enable pan
        self._toolbar.pan()

    # Activates the Axis-Limits to Rectangle option
    def _axis_limits_to_rectangle(self, *args):

        # Disable pan if it is enabled
        limits_option = self._menu_options['axis_limits_to_rectangle']
        pan_option = self._menu_options['pan']

        if pan_option.get() and limits_option.get():
            pan_option.set(False)

        # Enable axis-limits to rectangle (called zoom by MPL)
        self._toolbar.zoom()

    # Update the image scaling law to match the current menu_options value
    def _update_scale(self, *args):

        scale = self.menu_option('scale')

        # Get only the finite slice data (NaNs and infs will be masked and thus ignored)
        slice_data = self._image.get_array()

        if scale == 'linear':
            norm = mpl.colors.Normalize()

        elif scale == 'log':
            norm = mpl.colors.LogNorm()

        elif scale == 'squared':
            norm = PowerNorm(gamma=2.)

        elif scale == 'square root':
            norm = PowerNorm(gamma=0.5)

        else:
            self._menu_options['scale'].set('linear')
            raise ValueError('Unknown scale: {0}'.format(scale))

        scale_limits = self.menu_option('scale_limits')

        if scale_limits == 'min max':

            norm.vmin = np.ma.min(slice_data)
            norm.vmax = np.ma.max(slice_data)

            if isinstance(norm.vmin, np.ma.MaskedArray):
                norm.vmin = norm.vmin.compressed()[0]

            if isinstance(norm.vmax, np.ma.MaskedArray):
                norm.vmax = norm.vmax.compressed()[0]

        elif scale_limits == 'zscale':

            zscale_lim = zscale(slice_data)

            norm.vmin = zscale_lim[0]
            norm.vmax = zscale_lim[1]

        else:

            custom_scale = scale_limits.split(',')
            unknown_scale = False

            if len(custom_scale) == 2:

                try:
                    norm.vmin = float(custom_scale[0])
                    norm.vmax = float(custom_scale[1])

                except ValueError:
                    unknown_scale = True

            else:
                unknown_scale = True

            if unknown_scale:
                self._menu_options['scale_limits'].set('min max')
                raise ValueError('Unknown scale limits: {0}'.format(scale_limits))

        # If scaling minimum and maximum are the same value, scale to just less and just greater
        # (having equal values causes MPL errors)
        if norm.vmin == norm.vmax:
            norm.vmin -= 0.1
            norm.vmax += 0.1

        # Clamp down to first non-zero data value for scaling minimum of LogNorm and PowerNorm
        # (otherwise MPL will do it, and also issue a warning)
        if (norm.vmin < 0) and (('log' in scale) or ('square' in scale)):
            slice_data = slice_data.compressed()

            positive_idxs = np.where(slice_data > 0)
            if len(positive_idxs[0]) > 0:
                norm.vmin = slice_data[positive_idxs].min()
            else:
                norm.vmin = 10**-11
                norm.vmax = 10**-10

        # Use almost zero instead of exactly zero to avoid division by zero in MPL's `LogNorm.inverse`
        if (norm.vmin == 0) and ('log' in scale):
            norm.vmin = (norm.vmax - norm.vmin) / 100

        # Ensure scaling minimum is less than maximum once other adjustments finished, reverse if necessary
        if norm.vmin > norm.vmax:
            _norm_max = norm.vmax
            norm.vmax = norm.vmin
            norm.vmin = _norm_max

        # Freeze the image display temporarily because updating the norm below causes colorbar to twitch
        self._image_display_manage('freeze')

        # Update the norm. Calling set_norm() will cause issues when switching norms, because colorbar
        # still has the previous norm, and set_norm() will try to update its min/max without updating
        # the norm scale. Therefore we set the norm manually, then update the colormap, then call changed()
        # ourselves on the AxisImage to update it
        self._image.norm = norm
        self._update_colormap()
        self._image.changed()

        # Thaw the image display since it was frozen above
        self._image_display_manage('thaw')

    # Update the image rotation to match current menu_options value
    def _update_rotation(self, *args):

        menu_option = self._menu_options['rotation_angle']
        rotation_angle = menu_option.get()

        if rotation_angle % 90 != 0:
            menu_option.set(0)
            raise ValueError("Rotation angle '{0}'must be evenly divisible by 90".format(rotation_angle))

        self._draw_slice()

    # Updates the image orientation setting (orient via display dictionary or via storage order) to match
    # current menu_options value
    def _update_orientation(self, *args):

        menu_option = self._menu_options['orientation']
        orientation = menu_option.get()

        # Set orientation
        if orientation in ('display', 'storage'):

            # Do nothing if the only valid orientation is 'storage' as there are no valid display settings
            if (self.meta_data.display_settings is None) or (not self.meta_data.display_settings.valid):
                return

            self._draw_slice()

        else:
            menu_option.set('display')
            raise ValueError('Unknown orientation setting selected: {0}'.format(orientation))

    # Update the axis inversion to match current menu_options value
    def _update_invert_axis(self, *args):

        invert_axis = self.menu_option('invert_axis')
        axis = self._image.axes

        # Reset any axis inversion
        if axis.yaxis_inverted():
            axis.invert_yaxis()

        if axis.xaxis_inverted():
            axis.invert_xaxis()

        # Set axes inversion to proper orientation based on Display Settings
        display_settings = self.meta_data.display_settings
        use_display_orientation = self.menu_option('orientation') == 'display'

        if (display_settings is not None) and display_settings.valid and use_display_orientation:
            display_direction = display_settings['Display_Direction']

            if 'Top to Bottom' in display_direction['vertical_display_direction']:
                axis.invert_yaxis()

            if 'Right to Left' in display_direction['horizontal_display_direction']:
                axis.invert_xaxis()

        # Invert X, Y or both axes based on current menu settings
        if invert_axis == 'x':
            axis.invert_xaxis()

        elif invert_axis == 'y':
            axis.invert_yaxis()

        elif invert_axis == 'both':
            axis.invert_xaxis()
            axis.invert_yaxis()

        elif invert_axis != 'none':
            self._menu_options['invert_axis'].set('none')
            raise ValueError('Unknown axis to invert: {0}'.format(invert_axis))

        self._figure_canvas.draw()

    # Update the colormap to match the current menu_options value
    def _update_colormap(self, *args):

        cmap = self.menu_option('colormap')

        # Special handling for RGB images
        if self._settings['is_rgb']:

            if not self._data_open:
                return

            # Detect if change from RGB to non-RGB colormap occurred
            # (slice_data goes from 2D to 3D or vice-versa due to extra color axis being added/removed)
            current_slice = self._get_slice_data()
            new_slice = self._get_slice_data(update=True)

            if len(current_slice.shape) != len(new_slice.shape):

                # Update data
                self._slice_data = new_slice
                self._image.set_data(weakref.proxy(new_slice))

                # Update the scale (since the data changes)
                self._update_scale()

                # Do not continue (update scale will call this method again, and when it does the colormap
                # will be updated for monochrome images)
                return

            # Do not update colormap for RGB images
            elif cmap == 'RGB':
                return

        if self.menu_option('invert_colormap'):
            cmap += '_r'

        # Update the cmap. Calling set_cmap() will cause issues when also switching norms, because colorbar
        # still has the previous norm, and set_cmap() will try to update its min/max without updating the norm
        # type. Therefore we set the cmap manually, then update the colormap, then call changed() ourselves on
        # the AxisImage to update it
        self._image.cmap = mpl.cm.get_cmap(cmap)
        self._update_colorbar()
        self._image.changed()
        self._figure_canvas.draw()

    # Update the colorbar (e.g. if it's been changed from horizontal to vertical, the slice being viewed has
    # changed, or the image's colormap, scaling law, dimensions have been changed)
    def _update_colorbar(self, full_redraw=False, *args):

        if self.menu_option('show_colorbar') and (self._colorbar is not None):

            # Remove and re-draw if requested (e.g. on change of vertical/horizontal)
            if full_redraw:

                self._remove_colorbar()
                self._add_colorbar()

            # Update only colors, scaling and ticks otherwise
            else:

                self._colorbar.update(self._image)

                # The scrollable canvas dimensions change if colorbar dimensions change and need updating
                self._update_scrollable_canvas_dimensions()

    # Shows or hides axis ticks, tick labels and border around image for by menu_options
    def _update_tick_labels(self, *args):

        axis = self._image.axes

        # Determine axis tick label size appropriate for current image dimensions
        slice_data = self._get_slice_data()
        zoom_level = self.menu_option('zoom_level')

        min_dimension = min(slice_data.shape) * zoom_level
        tick_size = int(math.log(min_dimension / 11.) * 4 + 2)

        if tick_size < 2:
            tick_size = 2

        axis.tick_params(axis='both', which='major', labelsize=tick_size)

        # Hide currently shown X and Y ticks
        x_was_visible = axis.get_xaxis().get_visible()
        y_was_visible = axis.get_yaxis().get_visible()

        axis.get_xaxis().set_visible(False)
        axis.get_yaxis().set_visible(False)

        # Show X and Y axis tick labels if requested
        show_axis_ticks = self.menu_option('show_axis_ticks')

        if show_axis_ticks == 'x':
            axis.get_xaxis().set_visible(True)

        elif show_axis_ticks == 'y':
            axis.get_yaxis().set_visible(True)

        elif show_axis_ticks == 'both':
            axis.get_xaxis().set_visible(True)
            axis.get_yaxis().set_visible(True)

        elif show_axis_ticks != 'none':
            self._menu_options['show_axis_ticks'].set('none')
            raise ValueError('Unknown axis selected to display: {0}'.format(show_axis_ticks))

        x_is_visible = axis.get_xaxis().get_visible()
        y_is_visible = axis.get_yaxis().get_visible()

        # Adjust margins (for addition/removal of tick labels). We adjust both X and Y margins
        # even if only a single side has tick labels because a tick label close to the side of the image
        # will overflow into the other margin if we adjust only a single one
        pixel_margin = 20 * len(str(max(slice_data.shape)))
        change_margin = None

        # Add margin only if: (1) neither x or y already were visible (2) one of them is visible now
        if (not x_was_visible and not y_was_visible) and (x_is_visible or y_is_visible):
            change_margin = pixel_margin
            self._settings['pixel_margins'] = (pixel_margin, pixel_margin)

        # Remove margin only if: (1) either x or y were already visible (2) neither of them is visible now
        elif (x_was_visible or y_was_visible) and (not x_is_visible and not y_is_visible):
            change_margin = -1 * pixel_margin
            self._settings['pixel_margins'] = (0, 0)

        # Adjust margin if necessary (either both tick labels were removed, or at least one was added)
        if change_margin is not None:

            # Freeze the image display temporarily because updating the image's dimensions causes it to twitch
            self._image_display_manage('freeze')

            # Calculate new figure dimensions for the pixel margin calculated above
            x_pixels, y_pixels = self._figure_canvas.get_dimensions(type='pixels')
            fig_size = (x_pixels + 2 * change_margin, y_pixels + 2 * change_margin)
            bbox = self._get_bbox_with_margins(fig_size)

            # Expand/shrink image dimensions while keeping image the same size via adjusting the margin
            self._image.axes.set_position(bbox)
            self._figure_canvas.set_dimensions(fig_size, type='pixels')

            # The scrollable canvas dimensions change since image dimensions change and need updating
            self._update_scrollable_canvas_dimensions()

            # Thaw the image display since it was frozen above
            self._image_display_manage('thaw')

        # Show / hide image border if requested (works only when tick labels are also shown)
        if self.menu_option('show_border') and (x_is_visible or y_is_visible):
            axis.set_frame_on(True)

        else:
            axis.set_frame_on(False)

        self._figure_canvas.draw()

    # Update menu_options to specified zoom_level, and then call _update_zoom() to zoom to that level. If
    # no zoom level is specified then an action may be specified, and the image is zoomed in or out (depending
    # on the action) by the zoom_factor variable
    def _zoom(self, action='in', zoom_level=None):

        # Determine zoom level if none if specified
        if zoom_level is None:

            # Obtain the current zoom level
            zoom_level = self.menu_option('zoom_level')

            # Determine zoom level when adjusted by zoom factor
            zoom_factor = 0.20

            if action == 'in':
                zoom_level *= (1 + zoom_factor)
            else:
                zoom_level *= (1 - zoom_factor)

        # Save the new zoom level
        self.set_zoom(zoom_level)

    # Zoom image to the level set by menu_options
    def _update_zoom(self, *args):

        # Freeze the image display temporarily so it is not needlessly re-drawn multiple times
        self._image_display_manage('freeze')

        fig_size = self._settings['pixel_init_dimensions']
        margins = self._settings['pixel_margins']

        # Determine what level of zoom to make
        zoom_level = self.menu_option('zoom_level')

        # Change figure size by zoom_level, taking care to prevent margins for tick labels from being
        # increased by zoom_level
        extra_space_width = margins[0] * 2
        extra_space_height = margins[1] * 2

        zoomed_size = [fig_size[0] * zoom_level + extra_space_width,
                       fig_size[1] * zoom_level + extra_space_height]

        # Ensure that the zoomed size is at least 1x1 pixel
        zoomed_size[0] = 1 if zoomed_size[0] < 1 else zoomed_size[0]
        zoomed_size[1] = 1 if zoomed_size[1] < 1 else zoomed_size[1]

        # Adjust margins of figure such that the image it contains will correspond to its zoomed image size
        bbox = self._get_bbox_with_margins(zoomed_size)
        self._image.axes.set_position(bbox)

        # Resize the figure to match the zoom level
        self._figure_canvas.set_dimensions(zoomed_size, type='pixels')

        # Update tick sizes match new zoom level
        self._update_tick_labels()

        # Thaw the image display since it was frozen above
        self._image_display_manage('thaw')

    # Sets the X, Y and pixel value in the header based on current mouse pointer location
    def _update_mouse_pixel_value(self, event):

        event_outside_bounds = True

        # If the event is an MPL MouseEvent then we may have scrolled mouse on the image
        if isinstance(event, mpl.backend_bases.MouseEvent):

            slice_data = self._get_slice_data()

            event_has_position = (event.xdata is not None) and (event.ydata is not None)
            is_on_image = event.inaxes == self._image.axes

            if event_has_position and is_on_image:

                is_in_x_bounds = (event.xdata >= -0.5) and (slice_data.shape[1] > (event.xdata + 0.5))
                is_in_y_bounds = (event.ydata >= -0.5) and (slice_data.shape[0] > (event.ydata + 0.5))

                if is_in_x_bounds and is_in_y_bounds:

                    x_position = event.xdata + 0.5
                    y_position = event.ydata + 0.5
                    pixel_value = slice_data[int(y_position), int(x_position)]

                    self._header_widgets['x'].set('{0:.01f}'.format(x_position))
                    self._header_widgets['y'].set('{0:.01f}'.format(y_position))
                    self._header_widgets['value'].set('{0:10}'.format(pixel_value).lstrip())

                    event_outside_bounds = False

        # Remove values from X, Y and pixel value if the mouse pointer is not on the image
        if event_outside_bounds:
            self._header_widgets['x'].set('')
            self._header_widgets['y'].set('')
            self._header_widgets['value'].set('')

    # Manage whether image seen on screen by user is frozen or not (this simulates freeze() and thaw()
    # available in WX GUI toolkit). If image is frozen then it does not update until it is thawed
    # (note that you must thaw it prior to exiting the function where it is frozen, otherwise the image will
    # simply disappear). If lock is True, any further calls to this function will be ignored unless they
    # specify lock is False.
    def _image_display_manage(self, action='thaw', lock=None):

        if (not self._image_display_locked) or (lock is False):

            # Hide or show display frame (this has the effect of freezing the image on the screen until
            # the next time the TKinter event loop is idle)
            if action == 'thaw':

                self._display_frame.pack(expand=1, fill='both')

                # Update scrollable canvas to account for new scrollregion needed
                self._update_scrollable_canvas_dimensions()

                # Thaw the image now that its being displayed again
                if self._figure_canvas is not None:
                    self._figure_canvas.thaw()

            elif action == 'freeze':

                if self._figure_canvas is not None:
                    self._figure_canvas.freeze()

                self._display_frame.pack_forget()

        if lock is not None:
            self._image_display_locked = lock

    # Returns an MPL bounding box such that the MPL axes having this bbox, inside a figure having fig_size,
    # will have margins equivalent to those in self._settings[]
    def _get_bbox_with_margins(self, fig_size):

        x_margin = self._settings['pixel_margins'][0] / fig_size[0]
        y_margin = self._settings['pixel_margins'][1] / fig_size[1]

        bbox = [x_margin, y_margin, 1. - 2. * x_margin, 1. - 2. * y_margin]

        return bbox

    # Called when the window has been resized, this method updates the dimensions of the various things that
    # need resizing together with a window resize
    def _window_resize(self, event):

        self._update_window_dimensions()
        self._update_scrollable_canvas_dimensions()

        if self._colorbar is not None:
            self._update_colorbar()

    # Updates scrollable canvas size and scroll region, needed any time the contents of scrollable canvas
    # have changed size
    def _update_scrollable_canvas_dimensions(self):

        self._widget.update_idletasks()

        # Retrieve width and height of the image
        dimensions = self._figure_canvas.get_dimensions()
        scroll_width = dimensions[0]
        scroll_height = dimensions[1]

        # Adjust width and height to account for the colorbar, which when present takes up space,
        # requiring that a bigger part of the image needs to be scrollable
        if self._colorbar is not None:

            if self._colorbar.orientation == 'vertical':
                scroll_width += self._colorbar.dimensions[0]

            else:
                scroll_height += self._colorbar.dimensions[1]

        # Adjust width and height of the _scrollable_canvas, while ensuring that it is not larger than the
        # window dimensions (or more specifically the static_canvas dimensions, because both header and
        # scrollbars take up some space)
        width = scroll_width
        if width > self._static_canvas.winfo_width():
            width = self._static_canvas.winfo_width()

        height = scroll_height
        if height > self._static_canvas.winfo_height():
            height = self._static_canvas.winfo_height()

        self._scrollable_canvas.configure(width=width, height=height)
        self._scrollable_canvas.configure(scrollregion=(0, 0, scroll_width, scroll_height))

    # Called on mouse wheel scroll action, zooms image in or out
    def _mousewheel_scroll(self, event):

        # A forced update seems to be required on Mac in order to not break scrollbars during
        # quick changes, e.g. on quick zoom changes via mousewheel.
        if platform.system() == 'Darwin':
            self._widget.update()

        # Zoom in
        if event.delta > 0:
            self._zoom('in')

        # Zoom out
        else:
            self._zoom('out')

    # Dialog window to save the image as currently displayed
    def _save_file_box(self):

        mpl_canvas = self._figure_canvas.mpl_canvas
        filetypes = mpl_canvas.get_supported_filetypes().copy()
        default_filetype = mpl_canvas.get_default_filetype()

        default_filetype_name = filetypes[default_filetype]
        del filetypes[default_filetype]

        sorted_filetypes = list(six.iteritems(filetypes))
        sorted_filetypes.sort()
        sorted_filetypes.insert(0, (default_filetype, default_filetype_name))

        tk_filetypes = [(name, '*.%s' % ext) for (ext, name) in sorted_filetypes]

        initial_dir = self._viewer.last_open_dir
        initial_file = mpl_canvas.get_default_filename()

        # We add an empty defaultextension because otherwise changing to any extension does not work properly
        filename = asksaveasfilename(title='Save the figure',
                                     parent=self._widget,
                                     filetypes=tk_filetypes,
                                     initialdir=initial_dir,
                                     initialfile=initial_file,
                                     defaultextension='')

        if filename == '' or filename == ():
            return

        self._viewer.last_open_dir = os.path.dirname(filename)

        # Save figure, ensure facecolor is not ignored
        mpl_canvas.print_figure(filename,
                                dpi=self._settings['dpi'],
                                facecolor=self._figure_canvas.mpl_figure.get_facecolor())

    def close(self):

        if self._image is not None:
            self._image.remove()

        if self._toolbar is not None:
            self._toolbar = None

        if self._figure_canvas is not None:
            self._figure_canvas.destroy()

        if self._colorbar is not None:
            self._colorbar.destroy()

        super(ImageViewWindow, self).close()


class Colorbar(object):
    """ Colorbar class; creates a FigureCanvas for the Colorbar and controls its content """

    def __init__(self, image_static_canvas, image, orientation):

        # Instance variables
        self._image_static_canvas = image_static_canvas
        self._image = image
        self._orientation = orientation

        self._frame = Frame(image_static_canvas)
        self._figure_canvas = None
        self._mpl_colorbar = None
        self._dimensions = None

        # Create the MPL Figure for the colorbar
        colorbar_figure = Figure(facecolor='#F0F0F0')

        # Setup a new axis for the external colorbar
        if orientation == 'vertical':
            self._frame.pack(side='right', fill='y')
            cax = colorbar_figure.add_axes([0.1, 0.05, 0.2, 0.9])

        else:
            self._frame.pack(side='bottom', fill='x')
            cax = colorbar_figure.add_axes([0.04, 0.4, 0.92, 0.4])

        # Create a new FigureCanvas for the Figure
        self._orientation = orientation
        self._figure_canvas = FigureCanvas(colorbar_figure, master=self._frame)
        self._mpl_colorbar = colorbar_figure.colorbar(image, orientation=orientation, cax=cax, format='%4.2e')

        # Update colorbar dimensions (in this case setting the colorbar's initial size)
        self._update_dimensions()

        # Pack and draw the figure canvas
        self._figure_canvas.pack()
        self._figure_canvas.draw()

    @property
    def orientation(self):
        return self._orientation

    @property
    def dimensions(self):
        return self._dimensions

    def destroy(self):

        # Remove the colorbar (MPL v1.3 and less do not have .remove method for it)
        try:
            self._mpl_colorbar.remove()
        except AttributeError:
            self._image.colorbar = None

        self._figure_canvas.destroy()
        self._frame.destroy()

    def update(self, image):

        self._figure_canvas.freeze()

        self._image = image
        self._update_dimensions()
        self._mpl_colorbar.set_norm(image.norm)
        self._mpl_colorbar.set_cmap(image.cmap)
        self._mpl_colorbar.update_normal(image)

        self._figure_canvas.thaw()

    def _update_dimensions(self):

        dpi = self._figure_canvas.dpi

        if self._orientation == 'vertical':

            canvas_frame_height = self._image_static_canvas.winfo_height()
            colorbar_size = (1.5 * dpi, canvas_frame_height)

        else:

            canvas_frame_width = self._image_static_canvas.winfo_width()
            colorbar_size = (canvas_frame_width, 0.7 * dpi)

        # Update colorbar dimensions
        self._figure_canvas.set_dimensions(colorbar_size, type='pixels')
        self._dimensions = colorbar_size

        # Update tick number to account for new dimensions
        self._mpl_colorbar.set_ticks(self._get_ticks())

        self._figure_canvas.draw()

    # Returns ticks for colorbar such that spacing between tick marks is equal and the ticks are able to fit
    # in the current window dimensions
    def _get_ticks(self):

        # Obtain the data for which to get the ticks
        array = self._image.get_array()

        norm = self._image.norm
        norm.autoscale_None(array)

        colorbar_size_inches = np.asarray(self.dimensions) / self._figure_canvas.dpi

        # Determine how many tick marks to have (double valued function)
        if self._orientation == 'vertical':
            num_ticks = int(1.5 * colorbar_size_inches[1])

            if num_ticks > 8:
                num_ticks = int(math.log(colorbar_size_inches[1], 1.2))

        else:
            num_ticks = int(0.8 * colorbar_size_inches[0])

            if num_ticks > 8:
                num_ticks = int(math.log(colorbar_size_inches[0], 1.35))

        # Make ticks at even spacial intervals
        edge_remove = 0.12
        ticks = []

        if num_ticks == 1:
            ticks.append(norm.inverse(0.5))

        elif num_ticks > 1:

            increment = (1. / (num_ticks - 1)) * (1. - edge_remove)

            for i in range(0, num_ticks):
                tick_norm = (edge_remove / 2.) + (increment * i)
                ticks.append(norm.inverse(tick_norm))

        return ticks


class DataCubeWindow(Window):
    """ Window used to show the data cube for an ImageViewWindow """

    def __init__(self, viewer, image_structure_window):

        # Set initial necessary variables and do other required initialization procedures
        super(DataCubeWindow, self).__init__(viewer)

        # Set the title
        self._set_title('{0} - Data Cube'.format(self._widget.title()))

        # Set the ImageViewWindow for this DataCube as an instance variable
        self._structure_window = image_structure_window

        self._selected_axis = IntVar()
        self._selected_axis.set(self._structure_window.settings['selected_axis'])
        self._add_trace(self._selected_axis, 'w',
                        lambda *args: self._structure_window.select_slice(axis=self._selected_axis.get()))

        self._sliders = []

        # Get the header (containing title of structure)
        structure_name = self._structure_window.structure.id

        header_box = Frame(self._widget, bg=self.get_bg('gray'))
        header_box.pack(anchor='nw', expand=1, fill='x')

        w = Label(header_box, text=structure_name, bg=self.get_bg('gray'))
        w.pack(anchor='center', pady=(5, 0))

        # Create a row for each axes above 2D of ImageViewWindow
        self._sliders_box = Frame(self._widget, bg=self.get_bg('gray'))
        self._sliders_box.pack(anchor='nw', expand=1, fill='x')

        settings = self._structure_window.settings
        axes = settings['axes']

        # In case of a 2D array (e.g. horizontal and vertical axes), create an immovable slider
        num_axes = len(axes)
        has_color = False if axes.find('type', 'color') is None else True

        if (num_axes == 2) or (has_color and num_axes == 3):
            self._create_axis_row(0, '', 0, 0, 0, False)

        # Create sliders for many-dimensional arrays
        else:

            excluded_axes = ('color', 'vertical', 'horizontal')
            slider_axes = [axes[i] for i in range(0, num_axes) if axes[i]['type'] not in excluded_axes]

            # Determine if we need selectors for axes (necessary for cases when there are multiple sliders)
            add_selectors = False
            if len(slider_axes) > 1:
                add_selectors = True

            for i, axis in enumerate(slider_axes):
                self._create_axis_row(i, axis['name'], axis['sequence_number'],
                                      axis['slice'], axis['length'] - 1, add_selectors)

        self._show_window()

    # Creates a row to allow changing of slice index for a single axis. Contains a slider for each axis,
    # and if add_selectors is True then a radio button to select the axis
    def _create_axis_row(self, slider_index, axis_name, axis_sequence, axis_slice, max_slice, add_selectors):

            slider_var = IntVar()
            slider_var.set(axis_slice)
            self._add_trace(slider_var, 'w', lambda *args: self._slider_moved(slider_index, axis_sequence))

            slider_row_box = Frame(self._sliders_box, bg=self.get_bg('gray'))
            slider_row_box.pack(pady=(20, 10), expand=1, fill='x')

            # Add axis name and entry for slice index number
            w = Label(slider_row_box, text=axis_name, bg=self.get_bg('gray'))
            w.grid(row=slider_index, column=2, sticky='nw')

            w = Entry(slider_row_box, width=5, justify='center', highlightthickness=0, bd=0, bg='white',
                      textvariable=slider_var)
            w.grid(row=slider_index, column=2)

            # Add selection for the axis
            if add_selectors:
                radiobutton = Radiobutton(slider_row_box, variable=self._selected_axis, value=slider_index)
                radiobutton.grid(row=slider_index+1, column=0)

            # Add minimum, maximum and slider to control selected slice index
            w = Label(slider_row_box, text=0, bg=self.get_bg('gray'))
            w.grid(row=slider_index+1, column=1, padx=(10, 5))

            slider = Scale(slider_row_box, from_=0, to=max_slice, length=300, width=15,
                           bg=self.get_bg('gray'), orient='horizontal', showvalue=0, variable=slider_var)

            # Have slider expand/contract on window expansion/contraction
            slider_row_box.grid_columnconfigure(2, weight=1)
            slider.grid(row=slider_index+1, column=2, sticky='nswe')

            w = Label(slider_row_box, text=max_slice, bg=self.get_bg('gray'))
            w.grid(row=slider_index+1, column=3, padx=(5, 10))

            self._sliders.append(slider)

    # Calls the ImageViewWindow to update which frame is displayed
    def _slider_moved(self, slider_index, axis_sequence):

        index = self._sliders[slider_index].get()
        self._structure_window.select_slice(axis=axis_sequence, index=index)

    def close(self):

        self._structure_window = None
        super(DataCubeWindow, self).close()


class ScaleParametersWindow(Window):

    def __init__(self, viewer, image_structure_window, image):

        # Set initial necessary variables and do other required initialization procedures
        super(ScaleParametersWindow, self).__init__(viewer)

        # Set the title
        self._set_title("{0} - Scale Parameters".format(self._widget.title()))

        # Set instance variables for passed in parameters
        self._structure_window = image_structure_window
        self._structure_image = image

        # Set initial window dimensions
        self._settings = {'pixel_init_dimensions': (520, 160), 'dpi': 80}

        # Create the menu options
        menu_options = [
            {'name': 'histogram_range', 'type': StringVar(),  'default': 'full range', 'trace': self._update_histogram_range},
            {'name': 'scope',           'type': StringVar(),  'default': 'local',      'trace': self._update_scale},
            {'name': 'scale',           'type': StringVar(),
             'default': self._structure_window.menu_option('scale'), 'trace': self._update_scale},
            {'name': 'scale_limits',    'type': StringVar(),
             'default': self._structure_window.menu_option('scale_limits'), 'trace': self._update_scale},
        ]

        for option in menu_options:

            var = option['type']
            self._menu_options[option['name']] = var

            self._add_trace(var, 'w', option['trace'], option['default'])

        # Add the menu
        self._add_menu()

        # Extract a flattened view of the data, removing any NaN and inf
        flat_data = self._structure_window.data.ravel()
        self._histogram_data = flat_data[np.isfinite(flat_data)]

        # Create variables to store current scale limits
        self._scale_min = DoubleVar()
        self._scale_min.set(self._structure_image.norm.vmin)

        self._scale_max = DoubleVar()
        self._scale_max.set(self._structure_image.norm.vmax)

        # Create variables to store displayed vertical line for scale_min and scale_max
        self._min_line = None
        self._max_line = None

        # Create instance variables for a FigureCanvas containing the histogram image, and
        # a frame to store said FigureCanvas in
        self._histogram_frame = None
        self._figure_canvas = None

        # Draw the main window content
        self._draw_content()

        # Add notify event for window resizing
        self._fit_to_content()
        self._show_window()
        self._histogram_frame.bind('<Configure>', self._window_resize)

    # Adds menu options used for manipulating the data display
    def _add_menu(self):

        # Create File Menu
        self._menu = Menu(self._widget)
        self._widget.config(menu=self._menu)

        file_menu = Menu(self._menu, tearoff=0)
        file_menu.add_command(label='Close', command=self.close)
        file_menu.add_command(label='Close All', command=self._viewer.quit)
        self._menu.add_cascade(label='File', menu=file_menu)

        # Add a Scale menu
        scale_menu = Menu(self._menu, tearoff=0)
        self._menu.add_cascade(label='Scale', menu=scale_menu)

        scale_types = ('linear', 'log', 'squared', 'square root')

        for scale in scale_types:

            # PowerNorm is an MPL feature available in MPL v1.4+ only
            disable_powernorm = {}

            if (PowerNorm is None) and (scale in ('squared', 'square root')):
                disable_powernorm = {'state': 'disabled'}

            scale_menu.add_checkbutton(label=scale.title(), onvalue=scale, offvalue=scale,
                                       variable=self._menu_options['scale'], **disable_powernorm)

        # Add a Limits menu
        limits_menu = Menu(self._menu, tearoff=0)
        self._menu.add_cascade(label='Limits', menu=limits_menu)

        scale_limits = ('min max', 'zscale')

        for limit in scale_limits:
            limits_menu.add_checkbutton(label=limit.title(), onvalue=limit, offvalue=limit,
                                        variable=self._menu_options['scale_limits'])

        # Add a Scope menu
        scope_menu = Menu(self._menu, tearoff=0)
        self._menu.add_cascade(label='Scope', menu=scope_menu)

        scope_options = ('global', 'local')

        for scope in scope_options:
            scope_menu.add_checkbutton(label=scope.title(), onvalue=scope, offvalue=scope,
                                       variable=self._menu_options['scope'])

        # Add a Graph menu
        graph_menu = Menu(self._menu, tearoff=0)
        self._menu.add_cascade(label='Graph', menu=graph_menu)

        range_options = ('full range', 'current range')

        for graph_range in range_options:
            graph_menu.add_checkbutton(label=graph_range.title(), onvalue=graph_range, offvalue=graph_range,
                                       variable=self._menu_options['histogram_range'])

    # Draws the main content of the window (including the histogram and the buttons)
    def _draw_content(self):

        # Add the histogram title
        w = Label(self._widget, text='Pixel Distribution', font=self.get_font(size=9))
        w.pack(side='top')

        # Add the frame containing the limits entries and the Apply and Close buttons
        bottom_frame = Frame(self._widget)
        bottom_frame.pack(side='bottom', fill='x', anchor='nw', pady=(15, 10))

        limits_frame = Frame(bottom_frame)
        limits_frame.pack(side='top', anchor='nw')

        w = Label(limits_frame, text='Limits')
        w.pack(side='left')

        w = Label(limits_frame, text='Low')
        w.pack(side='left', padx=(10, 5))

        e = Entry(limits_frame, width=12, bg='white', textvariable=self._scale_min)
        e.pack(side='left')

        w = Label(limits_frame, text='High')
        w.pack(side='left', padx=(10, 5))

        e = Entry(limits_frame, width=12, bg='white', textvariable=self._scale_max)
        e.pack(side='left')

        separator = Frame(bottom_frame, relief='raised', borderwidth=1, height=2, bg='gray')
        separator.pack(side='top', fill='x', expand=1, pady=(5, 5))

        buttons_frame = Frame(bottom_frame)
        buttons_frame.pack(side='top', fill='x', expand=1)
        buttons_frame.grid_columnconfigure(0, weight=1)
        buttons_frame.grid_columnconfigure(1, weight=1)
        button_params = {'width': 10, 'font': self.get_font(weight='bold')}

        apply_button = Button(buttons_frame, text='Apply', command=self._apply_limits, **button_params)
        apply_button.grid(row=0, column=0)

        close_button = Button(buttons_frame, text='Close', command=self.close, **button_params)
        close_button.grid(row=0, column=1)

        # Get basic parameters needed for histogram
        dpi = self._settings['dpi']
        initial_fig_size = self._settings['pixel_init_dimensions']
        fig_size = (initial_fig_size[0] / dpi, initial_fig_size[1] / dpi)

        self._histogram_frame = Frame(self._widget)
        self._histogram_frame.pack(side='top', fill='both', expand=1)
        histogram_bg = 'white' if platform.system() == 'Darwin' else self.get_bg()

        figure = Figure(figsize=fig_size, dpi=dpi, facecolor=histogram_bg)
        self._figure_canvas = FigureCanvas(figure, master=self._histogram_frame)
        ax = self._figure_canvas.mpl_figure.add_axes([0.1, 0.13, 0.8, 0.80])

        # Disable tick marks on top and right sides, and enable grid
        ax.grid(color='gray', linestyle=':', linewidth=1,  axis='y', which='both', alpha=0.45)
        ax.set_axis_bgcolor(histogram_bg)
        ax.tick_params(direction='out')
        ax.yaxis.set_ticks_position('left')
        ax.xaxis.set_ticks_position('bottom')
        ax.locator_params(axis='x', nbins=5)

        self._histogram = ax.hist(self._histogram_data, bins=100, log=True, color='black')

        # Set histogram's axis limits to span exactly the desired range
        self._update_histogram_range()

        self._figure_canvas.pack()
        self._figure_canvas.draw()

    # Applies scale limits to image
    def _apply_limits(self):

        scale_min = self._scale_min.get()
        scale_max = self._scale_max.get()

        scale_limits = '{0},{1}'.format(scale_min, scale_max)
        self._menu_options['scale_limits'].set(scale_limits)

    # Updates scale and scale limits in image
    def _update_scale(self, *args):

        # Update scale and limits
        scale = self.menu_option('scale')
        scale_limits = self.menu_option('scale_limits')
        scope = self.menu_option('scope')

        # If scope is global and scale limits are not manual then we need to calculate scale limits
        # for all data and not just the slice being displayed as `set_scale` normally does
        if (scope == 'global') and (',' not in scale_limits):

            if scale_limits == 'zscale':
                zscale_lim = zscale(self._structure_window.data)
                min_x = zscale_lim[0]
                max_x = zscale_lim[1]

            else:
                min_x = self._histogram_data.min()
                max_x = self._histogram_data.max()

            scale_limits = '{0},{1}'.format(min_x, max_x)

        self._structure_window.set_scale(scale=scale, limits=scale_limits)

        # Update displayed limits to match image
        norm = self._structure_image.norm
        self._scale_min.set(norm.vmin)
        self._scale_max.set(norm.vmax)

        # Update histogram range to match (if necessary)
        self._update_histogram_range()

    # Updates histogram X-axis range (either show full range, or current range), and sets Y-axis range to
    # min and max.
    def _update_histogram_range(self, *args):

        histogram_range = self.menu_option('histogram_range')
        histogram_axis = self._figure_canvas.mpl_figure.axes[0]

        # Retrieve currently set scale limits for image/slice
        scale_min_set = self._scale_min.get()
        scale_max_set = self._scale_max.get()

        # Retrieve new X-axis limits for histogram
        if histogram_range == 'full range':
            scale_min = self._histogram_data.min()
            scale_max = self._histogram_data.max()

        elif histogram_range == 'current range':
            scale_min = scale_min_set
            scale_max = scale_max_set

        else:
            raise ValueError('Unknown histogram range: {0}'.format(histogram_range))

        # Do not continue if min and max of axes are equal to the same value, as this will cause an MPL error
        if scale_min == scale_max:
            return

        # Set X-axis limits of histogram to selected range
        # (1% more expansive than needed to exactly fit the data)
        x_min = scale_min - abs((scale_max - scale_min) * 0.01)
        x_max = scale_max + abs((scale_max - scale_min) * 0.01)
        histogram_axis.set_xlim(x_min, x_max)

        # Set X-axis limits of histogram to have 5 ticks at evenly spaced intervals
        interval = (scale_max - scale_min) / 4
        ticks = [scale_min] + [(scale_min + interval*i) for i in range(1, 4)] + [scale_max]
        histogram_axis.set_xticks(ticks)

        # Add vertical lines to histogram signifying currently set scale limits for image/slice
        if self._min_line is None:
            self._min_line = histogram_axis.axvline(x=scale_min_set, color='r', linewidth=2)
        else:
            self._min_line.set_xdata(scale_min_set)

        if self._max_line is None:
            self._max_line = histogram_axis.axvline(x=scale_max_set, color='g', linewidth=2)
        else:
            self._max_line.set_xdata(scale_max_set)

        # Set y-axis limits to span exactly the range of the data
        histogram_axis.set_ylim(self._histogram[0].min(), self._histogram[0].max())

        self._figure_canvas.draw()

    def _window_resize(self, event):

        self._update_window_dimensions()

        fig_size = (self._histogram_frame.winfo_width(), self._histogram_frame.winfo_height())
        self._figure_canvas.set_dimensions(fig_size, type='pixels')

    def close(self):

        self._figure_canvas.destroy()
        self._structure_window = None
        self._structure_image = None
        self._histogram_data = None

        super(ScaleParametersWindow, self).close()


class _AxesProperties(object):
    """ Helper class containing data about axes being displayed """

    def __init__(self):

        self.axes_properties = []

    def __getitem__(self, index):

        items = self.axes_properties[index]

        return items

    def __len__(self):
        return len(self.axes_properties)

    def add_axis(self, name, type, sequence_number, slice, length):

        axis_properties = {'name': name,
                           'type': type,
                           'sequence_number': sequence_number,
                           'slice': slice,
                           'length': length}

        self.axes_properties.append(axis_properties)

    # Finds an axis by property key and value
    def find(self, key, value):

        match = next((d for d in self.axes_properties if d[key] == value), None)

        if match is not None:
            match = match.copy()

        return match

    # Finds the index of an axis by property key and value
    def find_index(self, key, value):

        match = next((i for i, d in enumerate(self.axes_properties) if d[key] == value), None)
        return match

    # For axis having index, sets a key and a value
    def set(self, index, key, value):
        self.axes_properties[index][key] = value

    def copy(self):

        axes_properties = _AxesProperties()

        for axis in self.axes_properties:
            axes_properties.add_axis(axis['name'],
                                     axis['type'],
                                     axis['sequence_number'],
                                     axis['slice'],
                                     axis['length'])

        return axes_properties


# Opens a new ImageViewWindow from an ArrayStructure
def open_image(viewer, array_structure):

    image_window = ImageViewWindow(viewer)
    image_window.load_array(array_structure)

    return image_window
